from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from bson import ObjectId

from retk.models.tps.user import UserMeta


@dataclass
class AuthedUser:
    __slots__ = ("u", "language", "request_id")

    @dataclass
    class User:
        __slots__ = (
            "_id", "id", "source", "account", "nickname", "email", "avatar", "hashed", "disabled",
            "modified_at", "used_space", "type", "last_state", "settings"
        )

        @dataclass
        class LastState:
            __slots__ = (
                "node_display_method", "node_display_sort_key",
                "recent_search", "recent_cursor_search_selected_nids"
            )
            node_display_method: int
            node_display_sort_key: str
            recent_search: List[str]
            recent_cursor_search_selected_nids: List[str]

        @dataclass
        class Settings:
            __slots__ = (
                "language", "theme", "editor_mode", "editor_font_size",
                "editor_code_theme", "editor_sep_right_width", "editor_side_current_tool_id"
            )
            language: str
            theme: str
            editor_mode: str
            editor_font_size: int
            editor_code_theme: str
            editor_sep_right_width: int
            editor_side_current_tool_id: str

        _id: ObjectId
        id: str
        source: int
        account: str
        nickname: str
        email: str
        avatar: str
        hashed: str
        disabled: bool
        modified_at: datetime
        used_space: int
        type: int
        last_state: LastState
        settings: Settings

    u: Optional[User]
    language: str
    request_id: str


def convert_user_dict_to_authed_user(u: UserMeta) -> AuthedUser.User:
    return AuthedUser.User(
        _id=u["_id"],
        id=u["id"],
        source=u["source"],
        account=u["account"],
        nickname=u["nickname"],
        email=u["email"],
        avatar=u["avatar"],
        hashed=u["hashed"],
        disabled=u["disabled"],
        modified_at=u["modifiedAt"],
        used_space=u["usedSpace"],
        type=u["type"],
        last_state=AuthedUser.User.LastState(
            node_display_method=u["lastState"]["nodeDisplayMethod"],
            node_display_sort_key=u["lastState"]["nodeDisplaySortKey"],
            recent_search=u["lastState"]["recentSearch"],
            recent_cursor_search_selected_nids=u["lastState"]["recentCursorSearchSelectedNIds"],
        ),
        settings=AuthedUser.User.Settings(
            language=u["settings"]["language"],
            theme=u["settings"]["theme"],
            editor_mode=u["settings"]["editorMode"],
            editor_font_size=u["settings"]["editorFontSize"],
            editor_code_theme=u["settings"]["editorCodeTheme"],
            editor_sep_right_width=u["settings"].get("editorSepRightWidth", 200),
            editor_side_current_tool_id=u["settings"].get("editorSideCurrentToolId", ""),
        ),
    )



import inspect
from typing import Sequence
from urllib.parse import urlparse

from fastapi import HTTPException

from retk import const
from retk.logger import logger
from retk.models.tps import AuthedUser


def is_allowed_mime_type(data_url, allowed_mime_types: Sequence[str]):
    try:
        result = urlparse(data_url)
        if result.scheme != 'data':
            return False

        media_type = result.path.split(';')[0]
        return media_type in allowed_mime_types
    except Exception:  # pylint: disable=broad-except
        return False


def json_exception(
        request_id: str,
        code: const.CodeEnum,
        language: str = const.LanguageEnum.EN.value,
        log_msg: str = None
) -> HTTPException:
    def get_parent_function_info() -> str:
        previous_frame = inspect.currentframe().f_back.f_back
        if previous_frame.f_code.co_name == maybe_raise_json_exception.__name__:
            previous_frame = previous_frame.f_back
        previous_f_code = previous_frame.f_code
        caller = previous_f_code.co_name
        previous_frame_file = previous_f_code.co_filename
        previous_frame_line = previous_f_code.co_firstlineno
        return f"caller='{caller}' in {previous_frame_file}:{previous_frame_line}"

    if log_msg is not None:
        # get parent function name
        one_line_log = log_msg.replace("\n", "\\n")
        if code == const.CodeEnum.EXPIRED_AUTH:
            ll = logger.info
        else:
            ll = logger.error
        ll(f"reqId='{request_id}' | {get_parent_function_info()} | err='{one_line_log}'")

    status_code = const.CODE2STATUS_CODE[code]
    if status_code == 500:
        one_line_log = const.get_msg_by_code(code, language).replace("\n", "\\n")
        logger.error(f"reqId='{request_id}' | {get_parent_function_info()} | err='{one_line_log}'")

    return HTTPException(
        status_code=status_code,
        detail={
            "code": code.value,
            "msg": const.get_msg_by_code(code, language),
            "requestId": request_id,
        }
    )


def maybe_raise_json_exception(
        au: AuthedUser,
        code: const.CodeEnum,
):
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=au.request_id,
            code=code,
            language=au.language,
        )



import email.header
from datetime import timedelta
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from textwrap import dedent
from typing import List, Tuple

import jwt

from retk import const, config, regex, utils
from retk.core import scheduler


class EmailServer:
    default_language = const.LanguageEnum.EN

    lang_subject = {
        const.LanguageEnum.EN.value: "Rethink: Security Code",
        const.LanguageEnum.ZH.value: "Rethink: 安全密码",
    }
    lang_content = {
        const.LanguageEnum.EN.value: dedent("""\
        Please use the following security code for your Rethink account {email}:
        <br><br>
        Security Code: <br><br>
        <strong style="font-size:26px;">{numbers}</strong>
        <br><br>
        Valid for {expire} minutes, please do not tell others to prevent personal information leakage.
        <br><br>
        If you did not request this code, you can safely ignore this email, \
        someone may have entered your email address by mistake.
        <br><br>
        Thank you!<br>
        Rethink Team
        """),
        const.LanguageEnum.ZH.value: dedent("""\
        请使用以下用于 Rethink 账户 {email} 的安全代码：
        <br><br>
        安全代码：<br><br>
        <strong style="font-size:26px;">{numbers}</strong>
        <br><br>
        有效期 {expire} 分钟，请勿告知他人，以防个人信息泄露。
        <br><br>
        若您并未要求此代码，可以安全地忽视此邮件，可能有人误输入了您的电子邮件地址。
        <br><br>
        谢谢！<br>
        Rethink 团队
        """),
    }

    def get_subject_content(self, recipient: str, numbers: str, expire: int, language: str) -> Tuple[str, str]:
        try:
            subject = self.lang_subject[language]
            content_temp = self.lang_content[language]
        except KeyError:
            subject = self.lang_subject[self.default_language]
            content_temp = self.lang_content[self.default_language]
        content = content_temp.format(email=utils.mask_email(recipient), numbers=numbers, expire=expire)
        return subject, content

    def send(self, recipient: str, numbers: str, expire: int, language: str) -> const.CodeEnum:
        subject, content = self.get_subject_content(
            recipient=recipient, numbers=numbers, expire=expire, language=language
        )

        return self._send(
            subject=subject,
            recipients=[recipient],
            html_message=content
        )

    @staticmethod
    def email_ok(email_addr: str) -> bool:
        if regex.EMAIL.fullmatch(email_addr) is None:
            return False
        return True

    def _send(self, recipients: List[str], subject: str, html_message: str) -> const.CodeEnum:
        for recipient in recipients:
            if not self.email_ok(recipient):
                return const.CodeEnum.INVALID_EMAIL
        conf = config.get_settings()
        msg = MIMEMultipart('alternative')
        msg['Subject'] = email.header.Header(subject, 'utf-8')
        msg['From'] = conf.RETHINK_EMAIL
        msg['To'] = ", ".join(recipients)
        html_body = MIMEText(html_message, 'html', 'utf-8')
        msg.attach(html_body)

        _, code = scheduler.run_once_now(
            job_id=f"send_email_{conf.RETHINK_EMAIL}_{recipients}_{subject}_{html_message}",
            func=scheduler.tasks.email.send,
            kwargs={"recipients": recipients, "subject": msg.as_string()},
        )
        return code


email_server = EmailServer()


def encode_number(number: str, expired_min: int) -> str:
    token = utils.jwt_encode(
        exp_delta=timedelta(minutes=expired_min),
        data={"code": number + config.get_settings().CAPTCHA_SALT}
    )
    return token


def verify_number(token: str, number_str: str) -> const.CodeEnum:
    code = const.CodeEnum.CAPTCHA_ERROR
    try:
        data = utils.jwt_decode(token)
        if data["code"] == number_str + config.get_settings().CAPTCHA_SALT:
            code = const.CodeEnum.OK
    except jwt.ExpiredSignatureError:
        code = const.CodeEnum.CAPTCHA_EXPIRED
    except (jwt.DecodeError, Exception):  # pylint: disable=broad-except
        code = const.CodeEnum.INVALID_AUTH
    return code



from typing import Tuple

from pydantic import Field, BaseModel


class LatestVersionResponse(BaseModel):
    requestId: str
    hasNewVersion: bool
    local: Tuple[int, int, int]
    remote: Tuple[int, int, int] = Field(default=(0, 0, 0))



from retk.models.coll import Collections

IS_MOTOR = True
try:
    from motor.motor_asyncio import AsyncIOMotorCollection
except ImportError:
    IS_MOTOR = False


async def remote_try_build_index(coll: Collections):
    # try creating index
    if not IS_MOTOR:
        return

    await users_coll(coll.users)
    await nodes_coll(coll.nodes)
    await import_data_coll(coll.import_data)
    await user_file_coll(coll.user_file)
    await user_behavior_coll(coll.user_behavior)
    await notice_manager_delivery_coll(coll.notice_manager_delivery)
    await notice_system_coll(coll.notice_system)


async def not_in_and_create_index(coll: "AsyncIOMotorCollection", index_info, keys: list, unique: bool) -> str:
    new_keys = []
    for k in keys:
        if isinstance(k, tuple):
            new_keys.append(f"{k[0]}_{k[1]}")
        elif isinstance(k, str):
            new_keys.append(f"{k}_1")
        else:
            raise ValueError(f"Invalid key: {k}")
    index = "_".join(new_keys)
    if index not in index_info:
        await coll.create_index(keys, unique=unique)
    return index


async def users_coll(coll: "AsyncIOMotorCollection"):
    index_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["id"],
        unique=True
    )
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["account", "source"],
        unique=True
    )


async def nodes_coll(coll: "AsyncIOMotorCollection"):
    index_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["id"],
        unique=True
    )
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=[("uid", 1), ("id", -1)],
        unique=True
    )


async def import_data_coll(coll: "AsyncIOMotorCollection"):
    import_data_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=import_data_info,
        keys=["uid"],
        unique=True
    )


async def user_file_coll(coll: "AsyncIOMotorCollection"):
    user_file_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=user_file_info,
        keys=[("uid", 1), ("fid", -1)],
        unique=True
    )


async def user_behavior_coll(coll: "AsyncIOMotorCollection"):
    user_behavior_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=user_behavior_info,
        keys=["uid", "type"],
        unique=False
    )


async def notice_manager_delivery_coll(coll: "AsyncIOMotorCollection"):
    index_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["senderId"],
        unique=False
    )
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["scheduled"],
        unique=False
    )


async def notice_system_coll(coll: "AsyncIOMotorCollection"):
    index_info = await coll.index_information()
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["recipientId", "read"],
        unique=False
    )
    await not_in_and_create_index(
        coll=coll,
        index_info=index_info,
        keys=["senderId"],
        unique=False
    )



from .errors import MongitaNotImplementedError


class WriteConcern():
    def __init__(self, w=None, wtimeout=None, j=None, fsync=None):
        if any(p is not None for p in (w, wtimeout, j, fsync)):
            raise MongitaNotImplementedError("Mongita's WriteConcern is a dummy "
                                             "doesn't support parameters")

        self.w = None
        self.wtimeout = None
        self.j = None
        self.fsync = None

        self.acknowledged = True
        self.is_server_default = True

    @property
    def document(self):
        return {}



import datetime
from typing import Optional, Tuple

from bson import ObjectId
from bson.tz_util import utc

from retk import const, config
from retk.models.client import client
from retk.models.tps import ImportData


async def check_last_task_finished(uid: str, type_: str) -> Tuple[Optional[ImportData], bool]:
    doc = await client.coll.import_data.find_one({"uid": uid})
    if doc and doc["running"]:
        await set_running_false(
            uid=uid,
            code=const.CodeEnum.IMPORT_PROCESS_NOT_FINISHED,
            msg="last importData process not finished, please try again later",
        )
        return None, False

    if doc is None:
        doc: ImportData = {
            "_id": ObjectId(),
            "uid": uid,
            "process": 0,
            "type": type_,
            "startAt": datetime.datetime.now(tz=utc),
            "running": True,
            "msg": "",
            "code": 0,
            "obsidian": {},
        }
        res = await client.coll.import_data.insert_one(doc)
        if not res.acknowledged:
            await set_running_false(
                uid,
                const.CodeEnum.OPERATION_FAILED,
                msg="insert new importData process failed",
            )
            return doc, False
    else:
        doc, code = await update_process(
            uid=uid,
            type_=type_,
            process=0,
            start_at=datetime.datetime.now(tz=utc),
            running=True,
            code=0,
        )
        if code != const.CodeEnum.OK:
            await set_running_false(
                uid=uid,
                code=code,
                msg="update importData process failed",
            )
            return doc, False
    return doc, True


async def finish_task(uid: str, obsidian=None):
    resp = await client.coll.import_data.update_one(
        {"uid": uid},
        {"$set": {
            "obsidian": obsidian or {},
            "running": False,
            "process": 100,
            "code": 0,
            "msg": "done",
        }})

    if resp.modified_count != 1:
        await set_running_false(
            uid=uid,
            code=const.CodeEnum.OPERATION_FAILED,
            msg="uploading importData process failed",
        )


async def update_process(
        uid: str,
        type_: str,
        process: int,
        start_at: datetime.datetime = None,
        running: bool = None,
        code: int = None,
) -> Tuple[Optional[ImportData], const.CodeEnum]:
    data = {"type": type_, "process": process}
    if start_at is not None:
        data["startAt"] = start_at
    if running is not None:
        data["running"] = running
    if code is not None:
        data["code"] = code
    if config.is_local_db():
        # local db not support find_one_and_update
        await client.coll.import_data.update_one({"uid": uid}, {"$set": data})
        doc = await client.coll.import_data.find_one({"uid": uid})
    else:
        doc = await client.coll.import_data.find_one_and_update(
            {"uid": uid},
            {"$set": data}
        )
    if doc is None:
        return doc, const.CodeEnum.OPERATION_FAILED
    return doc, const.CodeEnum.OK


async def import_set_modules():
    from retk.models.client import client
    await client.init()


async def set_running_false(
        uid: str,
        code: const.CodeEnum,
        msg: str = "",
) -> None:
    await client.coll.import_data.update_one({"uid": uid}, {"$set": {
        "running": False,
        "msg": msg,
        "code": code.value,
    }})



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel, NonNegativeInt, Field, NonNegativeFloat

from retk import config
from retk.const import settings, LanguageEnum, app, NodeDisplaySortKeyEnum, USER_TYPE
from retk.models import tps
from retk.utils import datetime2str


class UserInfoResponse(BaseModel):
    class User(BaseModel):
        class LastState(BaseModel):
            nodeDisplayMethod: NonNegativeInt
            nodeDisplaySortKey: NodeDisplaySortKeyEnum

        class Settings(BaseModel):
            language: LanguageEnum
            theme: app.AppThemeEnum
            editorMode: app.EditorModeEnum
            editorFontSize: NonNegativeInt
            editorCodeTheme: app.EditorCodeThemeEnum
            editorSepRightWidth: float
            editorSideCurrentToolId: str

        email: str
        nickname: str
        avatar: str
        source: int
        createdAt: str
        usedSpace: NonNegativeInt = 0
        maxSpace: NonNegativeInt = 0
        lastState: LastState
        settings: Settings

    requestId: str
    user: User = None


class PatchUserRequest(BaseModel):
    class LastState(BaseModel):
        nodeDisplayMethod: Optional[NonNegativeInt] = Field(default=None, ge=-1, le=10)
        nodeDisplaySortKey: Optional[NodeDisplaySortKeyEnum] = Field(
            default=None, max_length=20
        )

    class Settings(BaseModel):
        language: Optional[LanguageEnum] = Field(default=None)
        theme: Optional[app.AppThemeEnum] = Field(default=None)
        editorMode: Optional[app.EditorModeEnum] = Field(default=None)
        editorFontSize: Optional[NonNegativeInt] = Field(default=None)
        editorCodeTheme: Optional[app.EditorCodeThemeEnum] = Field(default=None)
        editorSepRightWidth: Optional[NonNegativeFloat] = Field(default=None)
        editorSideCurrentToolId: Optional[str] = Field(default=None)

    nickname: Optional[str] = Field(default=None, max_length=settings.NICKNAME_MAX_LENGTH)
    avatar: Optional[str] = Field(default=None, max_length=2048)
    lastState: Optional[LastState] = Field(default=None)
    settings: Optional[Settings] = Field(default=None)


class UpdatePasswordRequest(BaseModel):
    oldPassword: str = Field(max_length=settings.PASSWORD_MAX_LENGTH)
    newPassword: str = Field(max_length=settings.PASSWORD_MAX_LENGTH)


def get_user_info_response_from_u_dict(
        u: tps.UserMeta,
        request_id: str,
) -> UserInfoResponse:
    if config.is_local_db():
        max_space = 0
    else:
        max_space = USER_TYPE.id2config(u["type"]).max_store_space
    last_state = u["lastState"]
    u_settings = u["settings"]
    return UserInfoResponse(
        requestId=request_id,
        user=UserInfoResponse.User(
            email=u["email"],
            nickname=u["nickname"],
            avatar=u["avatar"],
            source=u["source"],
            createdAt=datetime2str(u["_id"].generation_time),
            usedSpace=u["usedSpace"],
            maxSpace=max_space,
            lastState=UserInfoResponse.User.LastState(
                nodeDisplayMethod=last_state["nodeDisplayMethod"],
                nodeDisplaySortKey=last_state["nodeDisplaySortKey"],
            ),
            settings=UserInfoResponse.User.Settings(
                language=u_settings["language"],
                theme=u_settings["theme"],
                editorMode=u_settings["editorMode"],
                editorFontSize=u_settings["editorFontSize"],
                editorCodeTheme=u_settings["editorCodeTheme"],
                editorSepRightWidth=u_settings.get("editorSepRightWidth", 200),
                editorSideCurrentToolId=u_settings.get("editorSideCurrentToolId", ""),
            ),
        ),
    )


class NotificationResponse(BaseModel):
    class System(BaseModel):
        class Notice(BaseModel):
            id: str
            title: str
            snippet: str
            publishAt: str
            read: bool
            readTime: Optional[datetime]

        total: int
        notices: List[Notice]

    requestId: str
    hasUnread: bool
    system: System



from dataclasses import dataclass
from enum import Enum, auto


class Every(Enum):
    minute = 0
    hour = auto()
    day = auto()
    month = auto()


@dataclass
class Timing:
    at_second: int = 0
    at_minute: int = 0
    at_hour: int = 0
    at_day: int = 0
    every: Every = Every.hour


def __validate_time(day: int = None, hour: int = None, minute: int = None, second: int = None):
    if day is not None and (day >= 31 or day < 0):
        raise ValueError(f"Invalid day: {day}")
    if hour is not None and (hour >= 24 or hour < 0):
        raise ValueError(f"Invalid hour: {hour}")
    if minute is not None and (minute >= 60 or minute < 0):
        raise ValueError(f"Invalid minute: {minute}")
    if second is not None and (second >= 60 or second < 0):
        raise ValueError(f"Invalid second: {second}")


def every_minute_at(second: int = 0):
    __validate_time(second=second)
    return Timing(at_second=second, every=Every.minute)


def every_hour_at(minute: int, second: int = 0):
    __validate_time(minute=minute, second=second)
    return Timing(at_second=second, at_minute=minute, every=Every.hour)


def every_day_at(hour: int, minute: int = 0, second: int = 0):
    __validate_time(hour=hour, minute=minute, second=second)
    return Timing(at_second=second, at_minute=minute, at_hour=hour, every=Every.day)


def every_month_at(day: int, hour: int, minute: int = 0, second: int = 0):
    __validate_time(day=day, hour=hour, minute=minute, second=second)
    return Timing(at_second=second, at_minute=minute, at_hour=hour, at_day=day, every=Every.month)



from retk import const, core, config
from retk.controllers import schemas
from retk.controllers.utils import maybe_raise_json_exception
from retk.core import account, notice
from retk.core.user import reset_password
from retk.models.tps import AuthedUser
from retk.utils import mask_email, regex, datetime2str


async def get_user(
        au: AuthedUser,
) -> schemas.user.UserInfoResponse:
    _email = mask_email(au.u.email)
    if config.is_local_db():
        max_space = 0
    else:
        max_space = const.USER_TYPE.id2config(au.u.type).max_store_space
    return schemas.user.UserInfoResponse(
        requestId=au.request_id,
        user=schemas.user.UserInfoResponse.User(
            email=_email,
            nickname=au.u.nickname,
            avatar=au.u.avatar,
            source=au.u.source,
            createdAt=datetime2str(au.u._id.generation_time),
            usedSpace=au.u.used_space,
            maxSpace=max_space,
            lastState=schemas.user.UserInfoResponse.User.LastState(
                nodeDisplayMethod=au.u.last_state.node_display_method,
                nodeDisplaySortKey=au.u.last_state.node_display_sort_key,
            ),
            settings=schemas.user.UserInfoResponse.User.Settings(
                language=au.u.settings.language,
                theme=au.u.settings.theme,
                editorMode=au.u.settings.editor_mode,
                editorFontSize=au.u.settings.editor_font_size,
                editorCodeTheme=au.u.settings.editor_code_theme,
                editorSepRightWidth=au.u.settings.editor_sep_right_width,
                editorSideCurrentToolId=au.u.settings.editor_side_current_tool_id,
            ),
        ),
    )


async def patch_user(
        au: AuthedUser,
        req: schemas.user.PatchUserRequest,
) -> schemas.user.UserInfoResponse:
    u, code = await core.user.patch(
        au=au,
        req=req,
    )
    maybe_raise_json_exception(au=au, code=code)

    u["email"] = mask_email(u["email"])
    return schemas.user.get_user_info_response_from_u_dict(u=u, request_id=au.request_id)


async def update_password(
        au: AuthedUser,
        req: schemas.user.UpdatePasswordRequest
) -> schemas.RequestIdResponse:
    if regex.VALID_PASSWORD.match(req.newPassword) is None:
        return maybe_raise_json_exception(au=au, code=const.CodeEnum.INVALID_PASSWORD)

    ok = await account.manager.is_right_password(
        email=au.u.email,
        hashed=au.u.hashed,
        password=req.oldPassword,
    )
    if not ok:
        return maybe_raise_json_exception(au=au, code=const.CodeEnum.OLD_PASSWORD_ERROR)

    hashed = account.manager.hash_password(password=req.newPassword, email=au.u.email)
    code = await reset_password(uid=au.u.id, hashed=hashed)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def get_user_notices(
        au: AuthedUser,
        unread_only: bool,
        page: int,
        limit: int,
) -> schemas.user.NotificationResponse:
    nt, code = await notice.get_user_notices(au=au, unread_only=unread_only, page=page, limit=limit)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.user.NotificationResponse(
        requestId=au.request_id,
        hasUnread=nt["hasUnread"],
        system=schemas.user.NotificationResponse.System(
            total=nt["system"]["total"],
            notices=[
                schemas.user.NotificationResponse.System.Notice(
                    id=n["id"],
                    title=n["title"],
                    snippet=n["snippet"],
                    publishAt=n["publishAt"],
                    read=n["read"],
                    readTime=n["readTime"],
                ) for n in nt["system"]["notices"]],
        ),
    )


async def mark_system_notice_read(
        au: AuthedUser,
        notice_id: str,
) -> schemas.RequestIdResponse:
    code = await notice.mark_system_notice_read(uid=au.u.id, notice_id=notice_id)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def mark_all_system_notice_read(
        au: AuthedUser,
) -> schemas.RequestIdResponse:
    code = await notice.mark_all_system_notice_read(au=au)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )



from typing import List

import pymongo.errors

from retk import const, core
from retk.logger import logger
from retk.models.tps import AuthedUser, convert_user_dict_to_authed_user
from .. import utils


async def update_text_task(  # noqa: C901
        files: List[dict],
        max_file_size: int,
        uid: str,
        request_id: str,
):
    type_ = "md"
    await utils.import_set_modules()

    doc, finished = await utils.check_last_task_finished(uid=uid, type_=type_)
    if not finished:
        return

    for file in files:
        if not file["filename"].endswith(".md") and not file["filename"].endswith(".txt"):
            await utils.set_running_false(
                uid=uid,
                code=const.CodeEnum.INVALID_FILE_TYPE,
                msg=f"invalid file type: {file['filename']}",
            )
            return
        if file["size"] > max_file_size:
            await utils.set_running_false(
                uid=uid,
                code=const.CodeEnum.TOO_LARGE_FILE,
                msg=f"file size: {file['size']} > {max_file_size} (max file size): {file['filename']}",
            )
            return

    u, code = await core.user.get(uid=uid, disabled=False)
    au = AuthedUser(
        u=convert_user_dict_to_authed_user(u),
        request_id=request_id,
        language=u["settings"].get("language", const.LanguageEnum.EN.value),
    )
    for i, file in enumerate(files):
        try:
            md = file["content"].decode("utf-8")
        except (FileNotFoundError, OSError) as e:
            logger.error(f"error: {e}. filepath: {file['filename']}")
            await utils.set_running_false(
                uid=uid,
                code=const.CodeEnum.FILE_OPEN_ERROR,
                msg=f"file decode utf-8 failed, {e}: {file['filename']}",
            )
            return
        title = file["filename"].rsplit(".", 1)[0]
        md = title + "\n\n" + md
        try:
            _, code = await core.node.post(
                au=au,
                md=md,
                type_=const.NodeTypeEnum.MARKDOWN.value,
            )
        except pymongo.errors.DuplicateKeyError:
            continue
        except Exception as e:  # pylint: disable=broad-except
            logger.error(f"error: {e}. filepath: {file['filename']}")
            await utils.set_running_false(
                uid=uid,
                code=const.CodeEnum.FILE_OPEN_ERROR,
                msg=f"file decode utf-8 failed, {e}: {file['filename']}",
            )
            return
        if code != const.CodeEnum.OK:
            await utils.set_running_false(
                uid=uid,
                code=code,
                msg=f"file insert failed: {file['filename']}",
            )
            return
        if i % 20 == 0:
            doc, code = await utils.update_process(uid=uid, type_=type_, process=int(i / len(files) * 100))
            if code != const.CodeEnum.OK:
                await utils.set_running_false(
                    uid=uid,
                    code=code,
                    msg="uploading process update failed",
                )
                return
            if not doc["running"]:
                break

    await utils.finish_task(uid=uid)



from datetime import datetime

from bson import ObjectId
from bson.tz_util import utc

from retk import const, __version__
from retk.core import self_hosted, scheduler, user
from retk.core.notice import post_in_manager_delivery
from retk.logger import logger
from retk.models.client import client
from retk.models.tps import AuthedUser

_local_system_authed_user = AuthedUser(
    u=AuthedUser.User(
        _id=ObjectId(),
        id="system",
        source=0,
        account="system",
        nickname="system",
        email="system@rethink.com",
        avatar="",
        hashed="",
        disabled=False,
        modified_at=datetime.now(tz=utc),
        used_space=0,
        type=const.USER_TYPE.MANAGER.id,
        last_state=AuthedUser.User.LastState(
            node_display_method=0,
            node_display_sort_key="",
            recent_search=[],
            recent_cursor_search_selected_nids=[]
        ),
        settings=AuthedUser.User.Settings(
            language="en",
            theme="light",
            editor_mode="markdown",
            editor_font_size=14,
            editor_code_theme="default",
            editor_sep_right_width=300,
            editor_side_current_tool_id=""
        )
    ),
    language="en",
    request_id=""
)

__new_version_title_zh = "发现新版本 ({})"
__new_version_content_temp_zh = """
更新您的本地版本 ({}) 到最新 ({}) 以获取最新功能和错误修复。

请在终端中运行此命令来升级：
    
```
pip install -U retk
```
"""

__new_version_title_en = "Find new version ({})"
__new_version_content_temp_en = """
Update your local version ({}) to the latest version ({}) to get the latest features and bug fixes.

Please run this command in your terminal to update:

```
pip install -U retk
```
"""


async def notice_new_pkg_version():
    remote, code = await self_hosted.get_latest_pkg_version()
    if code != const.CodeEnum.OK:
        logger.error("get latest version failed")
        return
    local = self_hosted.parse_version(__version__)
    if local is None:
        logger.error("parse version failed")
        return

    has_new_version = False
    for vr, vl in zip(remote, local):
        if vr > vl:
            has_new_version = True
            break
    if has_new_version:
        res = await client.coll.notice_manager_delivery.find({
            "senderId": "system",
        }).to_list(None)
        u, code = await user.get_by_email(const.DEFAULT_USER["email"])
        if code != const.CodeEnum.OK:
            logger.error("get user failed")
            return
        language = u["settings"]["language"]

        local_version_str = ".".join(map(str, local))
        remote_version_str = ".".join(map(str, remote))

        title = (
            __new_version_title_zh if language == "zh" else __new_version_title_en
        ).format(local_version_str, remote_version_str)
        content = (
            __new_version_content_temp_zh if language == "zh" else __new_version_content_temp_en
        ).format(local_version_str, remote_version_str)
        for notice in res:
            if notice["title"] == __new_version_title_zh or notice["title"] == __new_version_title_en:
                return
        await post_in_manager_delivery(
            au=_local_system_authed_user,
            title=title,
            content=content,
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=datetime.now(tz=utc)
        )
        await scheduler.tasks.notice.async_deliver_unscheduled_system_notices()



from typing import List, Optional

from fastapi import APIRouter, UploadFile, Request

from retk.controllers import schemas
from retk.controllers.files import upload_files
from retk.routes import utils

router = APIRouter(
    prefix="/api/files",
    tags=["files"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/obsidian",
    status_code=202,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def upload_obsidian_files(
        au: utils.ANNOTATED_AUTHED_USER,
        files: List[UploadFile],
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await upload_files.upload_obsidian_files(
        au=au,
        files=files,
    )


@router.post(
    path="/text",
    status_code=202,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def upload_text_files(
        au: utils.ANNOTATED_AUTHED_USER,
        files: List[UploadFile],
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await upload_files.upload_text_files(
        au=au,
        files=files,
    )


@router.post(
    path="/vditor",
    status_code=200,
    response_model=schemas.files.VditorFilesResponse,
)
@utils.measure_time_spend
async def vditor_upload(
        au: utils.ANNOTATED_AUTHED_USER,
        req: Request,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.files.VditorFilesResponse:
    form = await req.form()
    file = form.get("file[]")
    return await upload_files.upload_file_vditor(
        au=au,
        file=file,
    )


@router.post(
    path="/vditor/images",
    status_code=200,
    response_model=schemas.files.VditorImagesResponse,
)
@utils.measure_time_spend
async def vditor_fetch_image(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.files.ImageVditorFetchRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.files.VditorImagesResponse:
    return await upload_files.fetch_image_vditor(
        au=au,
        req=req,
    )


@router.get(
    path="/upload-process",
    status_code=200,
    response_model=schemas.files.FileUploadProcessResponse,
)
@utils.measure_time_spend
async def get_upload_process(
        au: utils.ANNOTATED_AUTHED_USER,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.files.FileUploadProcessResponse:
    return await upload_files.get_upload_process(
        au=au,
    )



from typing import Optional

from fastapi import APIRouter, Cookie, Header
from fastapi.responses import JSONResponse

from retk import const
from retk.controllers import schemas, account
from retk.routes import utils

router = APIRouter(
    prefix="/api/account",
    tags=["user"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    "/",
    status_code=201,
)
@utils.measure_time_spend
async def signup(
        req_id: utils.ANNOTATED_REQUEST_ID,
        req: schemas.account.SignupRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> JSONResponse:
    return await account.signup(req_id=req_id, req=req)


@router.put(
    "/auto-login",
    status_code=200,
)
@utils.measure_time_spend
async def auto_login(
        token: str = Cookie(alias=const.settings.COOKIE_ACCESS_TOKEN, default=""),
        request_id: str = Header(
            default="", alias="RequestId", max_length=const.settings.MD_MAX_LENGTH
        )
) -> schemas.user.UserInfoResponse:
    return await account.auto_login(
        token=token,
        req_id=request_id,
    )


@router.put(
    "/login",
    status_code=200,
)
@utils.measure_time_spend
async def login(
        req_id: utils.ANNOTATED_REQUEST_ID,
        req: schemas.account.LoginRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> JSONResponse:
    return await account.login(req_id=req_id, req=req)


@router.put(
    path="/password",
    status_code=200,
)
@utils.measure_time_spend
async def forget_password(
        req_id: utils.ANNOTATED_REQUEST_ID,
        req: schemas.account.ForgetPasswordRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await account.forget(req_id=req_id, req=req)


@router.put(
    path="/email/send-code",
    status_code=200,
    response_model=schemas.account.TokenResponse,
)
@utils.measure_time_spend
async def email_verification(
        req_id: utils.ANNOTATED_REQUEST_ID,
        req: schemas.account.EmailVerificationRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.account.TokenResponse:
    return await account.email_send_code(req_id=req_id, req=req)


@router.get(
    "/access-token",
    status_code=200,
)
@utils.measure_time_spend
async def refresh_token(
        au: utils.ANNOTATED_REFRESH_TOKEN,  # check refresh token expiration
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> JSONResponse:
    return await account.get_new_access_token(au=au)


@router.put(
    "/logout",
    status_code=200,
)
@utils.measure_time_spend
async def logout(
        req_id: utils.ANNOTATED_REQUEST_ID,
        au: utils.ANNOTATED_AUTHED_USER,
) -> JSONResponse:
    return await account.logout(req_id=req_id, au=au)



from pydantic import BaseModel

# flake8: noqa
from . import (
    account,
    user,
    node,
    recent,
    files,
    plugin,
    app_system,
    manager,
    statistic,
    notice,
)


class RequestIdResponse(BaseModel):
    requestId: str



"""A generic OAuth client that can be used to quickly create support for any OAuth provider
with close to no code
"""

import logging
from typing import Any, Callable, Dict, List, Optional, Type, Union

from .base import DiscoveryDocument, OpenID, SSOBase

logger = logging.getLogger(__name__)


def create_provider(
        *,
        name: str = "generic",
        default_scope: Optional[List[str]] = None,
        discovery_document: Union[DiscoveryDocument, Callable[[SSOBase], DiscoveryDocument]],
        response_convertor: Optional[Callable[[Dict[str, Any]], OpenID]] = None
) -> Type[SSOBase]:
    """A factory to create a generic OAuth client usable with almost any OAuth provider.
    Returns a class.

    Args:

        name: Name of the provider
        default_scope: default list of scopes (can be overriden in constructor)
        discovery_document: a dictionary containing discovery document or a callable returning it
        response_convertor: a callable that will receive JSON response from the userinfo endpoint
            and should return OpenID object

    Example:
        ```python
        from fastapi_sso.sso.generic import create_provider

        discovery = {
            "authorization_endpoint": "http://localhost:9090/auth",
            "token_endpoint": "http://localhost:9090/token",
            "userinfo_endpoint": "http://localhost:9090/me",
        }

        SSOProvider = create_provider(name="oidc", discovery_document=discovery)
        sso = SSOProvider(
            client_id="test",
            client_secret="secret",
            redirect_uri="http://localhost:8080/callback",
            allow_insecure_http=True
            )
        ```

    """

    class GenericSSOProvider(SSOBase):
        """SSO Provider Template"""

        provider = name
        scope = default_scope or ["openid"]

        async def get_discovery_document(self) -> DiscoveryDocument:
            """Get document containing handy urls"""
            if callable(discovery_document):
                return discovery_document(self)
            return discovery_document

        @classmethod
        async def openid_from_response(cls, response: dict) -> OpenID:
            if not response_convertor:
                logger.warning("No response convertor was provided, returned OpenID will always be empty")
                return OpenID(
                    provider=cls.provider,
                )
            return response_convertor(response)

    return GenericSSOProvider



from typing import List

from retk import const, core
from retk.controllers import schemas
from retk.controllers.utils import maybe_raise_json_exception
from retk.models.tps import AuthedUser, Node
from retk.utils import contain_only_http_link, get_title_description_from_link, datetime2str


def __get_node_data(n: Node) -> schemas.node.NodeData:
    from_nodes: List[schemas.node.NodeData.LinkedNode] = []
    to_nodes: List[schemas.node.NodeData.LinkedNode] = []
    for nodes, n_nodes in zip(
            [from_nodes, to_nodes],
            [n.get("fromNodes", []), n.get("toNodes", [])]):
        for _n in n_nodes:
            nodes.append(
                schemas.node.NodeData.LinkedNode(
                    id=_n["id"],
                    title=_n["title"],
                    md=_n["md"],
                    snippet=_n["snippet"],
                    type=_n["type"],
                    disabled=_n["disabled"],
                    inTrash=_n["inTrash"],
                    createdAt=datetime2str(_n["_id"].generation_time),
                    modifiedAt=datetime2str(_n["modifiedAt"]),
                )
            )
    return schemas.node.NodeData(
        id=n["id"],
        md=n["md"],
        title=n["title"],
        snippet=n["snippet"],
        type=n["type"],
        disabled=n["disabled"],
        createdAt=datetime2str(n["_id"].generation_time),
        modifiedAt=datetime2str(n["modifiedAt"]),
        fromNodes=from_nodes,
        toNodes=to_nodes,
    )


async def post_node(
        au: AuthedUser,
        req: schemas.node.CreateRequest,
        is_quick: bool = False,
) -> schemas.node.NodeResponse:
    n, code = await core.node.post(
        au=au,
        md=req.md,
        type_=req.type,
        from_nid=req.fromNid,
    )
    maybe_raise_json_exception(au=au, code=code)
    b_type = const.UserBehaviorTypeEnum.NODE_CREATE if not is_quick else const.UserBehaviorTypeEnum.NODE_QUICK_CREATE
    await core.statistic.add_user_behavior(
        uid=au.u.id,
        type_=b_type,
        remark=n['id'],
    )
    return schemas.node.NodeResponse(
        requestId=au.request_id,
        node=__get_node_data(n),
    )


async def post_quick_node(
        au: AuthedUser,
        req: schemas.node.CreateRequest,
) -> schemas.node.NodeResponse:
    if contain_only_http_link(req.md) != "":
        title, description = await get_title_description_from_link(
            url=req.md,
            language=au.language,
        )
        if au.language == const.LanguageEnum.ZH.value:
            desc_prefix = "**描述：**\n\n"
            link_prefix = "*链接：*"
        elif au.language == const.LanguageEnum.EN.value:
            desc_prefix = "**Description:**\n\n"
            link_prefix = "*Link:* "
        else:
            desc_prefix = "**Description:**\n\n"
            link_prefix = "*Link:* "
        req.md = f"{title}\n\n{desc_prefix}{description}\n\n{link_prefix}[{req.md}]({req.md})"

    return await post_node(
        au=au,
        req=req,
        is_quick=True,
    )


async def get_node(
        au: AuthedUser,
        nid: str,
) -> schemas.node.NodeResponse:
    n, code = await core.node.get(au=au, nid=nid)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.node.NodeResponse(
        requestId=au.request_id,
        node=__get_node_data(n),
    )


async def update_md(
        au: AuthedUser,
        req: schemas.node.PatchMdRequest,
        nid: str,
) -> schemas.node.NodeResponse:
    n, _, code = await core.node.update_md(
        au=au,
        nid=nid,
        md=req.md,
    )
    maybe_raise_json_exception(au=au, code=code)

    return schemas.node.NodeResponse(
        requestId=au.request_id,
        node=__get_node_data(n),
    )


async def get_core_nodes(
        au: AuthedUser,
        p: int,
        limit: int,
) -> schemas.node.NodesSearchResponse:
    nodes, total = await core.node.core_nodes(
        au=au,
        page=p,
        limit=limit,
    )
    return schemas.node.NodesSearchResponse(
        requestId=au.request_id,
        data=_get_node_search_response_data(nodes=nodes, total=total),
    )


def _get_node_search_response_data(nodes: List[Node], total: int) -> schemas.node.NodesSearchResponse.Data:
    return schemas.node.NodesSearchResponse.Data(
        nodes=[
            schemas.node.NodesSearchResponse.Data.Node(
                id=n["id"],
                title=n["title"],
                snippet=n["snippet"],
                titleHighlight="",
                bodyHighlights=[],
                score=0,
                type=n["type"],
                createdAt=datetime2str(n["_id"].generation_time),
                modifiedAt=datetime2str(n["modifiedAt"]),
            ) for n in nodes],
        total=total,
    )


async def get_hist_editions(
        au: AuthedUser,
        nid: str,
) -> schemas.node.HistEditionsResponse:
    versions, code = await core.node.get_hist_editions(
        au=au,
        nid=nid,
    )
    maybe_raise_json_exception(au=au, code=code)

    return schemas.node.HistEditionsResponse(
        requestId=au.request_id,
        versions=versions,
    )


async def get_hist_edition_md(
        au: AuthedUser,
        nid: str,
        version: str,
) -> schemas.node.HistEditionMdResponse:
    md, code = await core.node.get_hist_edition_md(
        au=au,
        nid=nid,
        version=version,
    )
    maybe_raise_json_exception(au=au, code=code)

    return schemas.node.HistEditionMdResponse(
        requestId=au.request_id,
        md=md,
    )



import datetime
import ipaddress
import math
import os
import re
import socket
import uuid
import webbrowser
from html.parser import HTMLParser
from io import StringIO
from typing import Tuple, Optional, List, Dict, Any
from urllib.parse import urlparse

import httpx
import jwt
from bson import ObjectId
from markdown import Markdown

from retk import config, const, regex
from retk.logger import logger
from retk.models import tps

HEADERS = {
    'typ': 'jwt',
    'alg': 'RS256'
}
alphabet = "3467ACDEFGHJKLMNPQRTUVWXYabcdefghkmnoprtuvwxy"
alphabet_len = len(alphabet)
__padding = int(math.ceil(math.log(2 ** 128, alphabet_len)))


def short_uuid() -> str:
    """
    The output has the most significant digit first.
    """
    number = uuid.uuid4().int
    output = ""

    while number:
        number, digit = divmod(number, alphabet_len)
        output += alphabet[digit]
    if __padding:
        remainder = max(__padding - len(output), 0)
        output = output + alphabet[0] * remainder
    return output[::-1]


def jwt_encode(exp_delta: datetime.timedelta, data: dict) -> str:
    """
    Create token

    Args:
        exp_delta: expired delta
        data: payload data

    Returns:
        str: token
    """
    payload = {
        "exp": datetime.datetime.utcnow() + exp_delta
    }
    payload.update(data)
    token = jwt.encode(
        payload=payload,
        key=config.get_settings().JWT_KEY,
        algorithm=HEADERS["alg"],
        headers=HEADERS,
    )
    return f"Bearer {token}"


def jwt_decode(token: str) -> dict:
    t = token.split("Bearer ", maxsplit=1)
    if len(t) != 2:
        raise jwt.InvalidTokenError("Invalid token")
    return jwt.decode(
        t[1],
        key=config.get_settings().JWT_KEY_PUB,
        algorithms=[HEADERS["alg"]],
        options={"verify_exp": True}
    )


class HTMLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, d):
        self.text.write(d)

    def get_data(self):
        return self.text.getvalue()


def strip_html_tags(html: str) -> str:
    s = HTMLStripper()
    s.feed(html[:1000])
    return s.get_data()


def split_title_body(fulltext: str) -> (str, str):
    title_body = fulltext.split("\n", maxsplit=1)
    title = title_body[0].strip()
    try:
        body = title_body[1].strip()
    except IndexError:
        body = ""
    return title, body


def __unmark_element(element, stream=None):
    if stream is None:
        stream = StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        __unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


# patching Markdown
Markdown.output_formats["plain"] = __unmark_element
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False
__md_html = Markdown(
    output_format="html",
)


def md2txt(md: str) -> str:
    for found in list(regex.MD_CODE.finditer(md))[::-1]:
        span = found.span()
        code = found.group(1)
        md = f"{md[: span[0]]}{code}{md[span[1]:]}"
    return __md.convert(md)


def preprocess_md(md: str, snippet_len: int = 200) -> Tuple[str, str, str]:
    title, body = split_title_body(fulltext=md)
    title = md2txt(title.strip())
    body = md2txt(body.strip())
    snippet = strip_html_tags(body)[:snippet_len]
    return title, body, snippet


def md2html(md: str) -> str:
    _html = __md_html.convert(md)
    # prevent XSS and other security issues
    _html = re.sub(r"<script[^>]*>.*?</script>", "", _html, flags=re.DOTALL)
    return _html


def change_link_title(md: str, nid: str, new_title: str) -> str:
    new_md = re.sub(
        r"\[@[^].]*?]\(/n/{}/?\)".format(nid),
        f"[@{new_title}](/n/{nid})",
        md,
    )
    return new_md


def contain_only_http_link(md: str) -> str:
    content = md.strip()
    if regex.ONLY_HTTP_URL.match(content) is None:
        return ""
    return content


def datetime2str(dt: datetime.datetime) -> str:
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


ASYNC_CLIENT_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,"
              "application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh,en-GB;q=0.9,en-US;q=0.8,en;q=0.7",
    "Cache-Control": "no-cache",
    "Pragma": "no-cache",
    "Sec-Ch-Ua": '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": '"macOS"',
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "none",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 "
                  "(KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
}


def is_internal_ip(ip):
    internal_ip_ranges = [
        ("127.0.0.0", "127.255.255.255"),
        ("10.0.0.0", "10.255.255.255"),
        ("9.0.0.0", "9.255.255.255"),
        ("100.64.0.0", "100.127.255.255"),
        ("192.168.0.0", "192.168.255.255"),
        ("172.16.0.0", "172.31.255.255"),
    ]

    for start, end in internal_ip_ranges:
        if ipaddress.IPv4Address(start) <= ip <= ipaddress.IPv4Address(end):
            return True
    return False


# SSRF protection
def ssrf_check(url: str) -> bool:
    if url.startswith("http://localhost") or url.startswith("http://127.0.0.1"):
        return True
    if '@' in url:
        return True
    host = urlparse(url).hostname
    if host is None:
        return True
    # cos url
    settings = config.get_settings()
    if host == f"{settings.COS_BUCKET_NAME}.cos.{settings.COS_REGION}.myqcloud.com":
        return True
    try:
        ip = ipaddress.IPv4Address(socket.gethostbyname(host))
    except socket.gaierror:
        return True
    if ip.is_private:
        return True
    if is_internal_ip(ip):
        return True
    return False


async def get_title_description_from_link(  # noqa: C901
        url: str,
        language: str,
        count=0
) -> Tuple[str, str]:
    if language == const.LanguageEnum.ZH.value:
        no_title = "网址没发现标题"
        no_description = "网址没发现描述"
    elif language == const.LanguageEnum.EN.value:
        no_title = "No title found"
        no_description = "No description found"
    else:
        no_title = "No title found"
        no_description = "No description found"

    if count > 2:
        logger.debug(f"too many 30X code, failed to get {url}")
        return no_title, no_description

    # SSRF protection
    if ssrf_check(url):
        return no_title, no_description
    # end of SSRF protection

    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                url=url,
                headers=ASYNC_CLIENT_HEADERS,
                timeout=3.
            )
        except (
                httpx.ConnectTimeout,
                RuntimeError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            logger.debug(f"failed to get {url}: {e}")
            return no_title, no_description
        if response.status_code in (301, 302):
            return await get_title_description_from_link(
                url=response.headers["Location"],
                language=language,
                count=count + 1,
            )
        if response.status_code != 200:
            return no_title, no_description
        html = response.text

    title, description = "", ""
    found = re.search(r'<meta[^>]*name="title"[^>]*content="([^"]*)"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r'<meta[^>]*content="([^"]*)"[^>]*name="title"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r"<title[^>]*>(.*?)</title>", html, re.DOTALL)
    if found:
        title = found.group(1).strip()

    found = re.search(r'<meta[^>]*name="description"[^>]*content="([^"]*)"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r'<meta[^>]*content="([^"]*)"[^>]*name="description"[^>]*>', html, re.DOTALL)
    if found:
        description = found.group(1).strip()[:400]
    if title == "":
        title = no_title
    if description == "":
        description = no_description
    return title, description


def mask_email(email: str):
    if email == "":
        return ""
    if "@" not in email:
        return email
    name, end = email.split("@", 1)
    if len(name) <= 1:
        e = f"{name[0]}**@{end}"
    elif len(name) == 2:
        e = f"{name[0]}**{name[1]}@{end}"
    else:
        e = f"{name[:2]}**{name[-1]}@{end}"
    return e


def local_finish_up():
    if config.is_local_db():
        addr = os.getenv('VUE_APP_API_URL')
        print(f"Rethink running on {addr} (Press CTRL+C to quit)")
        if os.getenv("RETHINK_SERVER_HEADLESS", "0") == "0" and addr:
            webbrowser.open_new_tab(addr)


def get_user_dict(
        _id: ObjectId,
        uid: str,
        source: int,
        account: str,
        nickname: str,
        email: str,
        avatar: str,
        hashed: str,
        disabled: bool,
        modified_at: datetime.datetime,
        used_space: int,
        type_: int,

        last_state_recent_cursor_search_selected_nids: List[str],
        last_state_recent_search: List[str],
        last_state_node_display_method: int,
        last_state_node_display_sort_key: str,

        settings_language: str,
        settings_theme: str,
        settings_editor_mode: str,
        settings_editor_font_size: int,
        settings_editor_code_theme: str,
        settings_editor_sep_right_width: int,
        settings_editor_side_current_tool_id: str,
) -> tps.UserMeta:
    return {
        "_id": _id,
        "id": uid,
        "source": source,
        "account": account,
        "nickname": nickname,
        "email": email,
        "avatar": avatar,
        "hashed": hashed,
        "disabled": disabled,
        "modifiedAt": modified_at,
        "usedSpace": used_space,
        "type": type_,

        "lastState": {
            "recentCursorSearchSelectedNIds": last_state_recent_cursor_search_selected_nids,
            "recentSearch": last_state_recent_search,
            "nodeDisplayMethod": last_state_node_display_method,
            "nodeDisplaySortKey": last_state_node_display_sort_key,
        },
        "settings": {
            "language": settings_language,
            "theme": settings_theme,
            "editorMode": settings_editor_mode,
            "editorFontSize": settings_editor_font_size,
            "editorCodeTheme": settings_editor_code_theme,
            "editorSepRightWidth": settings_editor_sep_right_width,
            "editorSideCurrentToolId": settings_editor_side_current_tool_id,
        },
    }


def get_node_dict(
        _id: ObjectId,
        nid: str,
        uid: str,
        md: str,
        title: str,
        snippet: str,
        type_: int,
        disabled: bool,
        in_trash: bool,
        modified_at: datetime.datetime,
        in_trash_at: Optional[datetime.datetime],
        from_node_ids: List[str],
        to_node_ids: List[str],
        history: List[Dict[str, Any]],
) -> tps.Node:
    return {
        "_id": _id,
        "id": nid,
        "uid": uid,
        "md": md,
        "title": title,
        "snippet": snippet,
        "type": type_,
        "disabled": disabled,
        "inTrash": in_trash,
        "modifiedAt": modified_at,
        "inTrashAt": in_trash_at,
        "fromNodeIds": from_node_ids,
        "toNodeIds": to_node_ids,
        "history": history,
    }


def get_token(uid: str, language: str) -> Tuple[str, str]:
    settings = config.get_settings()
    access_token = jwt_encode(
        exp_delta=settings.ACCESS_TOKEN_EXPIRE_DELTA,
        data={
            "is_access": True,
            "uid": uid,
            "language": language,
        },
    )
    refresh_token = jwt_encode(
        exp_delta=settings.REFRESH_TOKEN_EXPIRE_DELTA,
        data={
            "is_access": False,
            "uid": uid,
            "language": language,
        },
    )
    return access_token, refresh_token



import json
from pathlib import Path
from typing import Union, Dict, Optional

from retk import const
from retk._version import __version__
from retk.logger import logger


def version_tuple(v):
    return tuple(map(int, (v.split("."))))


__current_version = version_tuple(__version__)


def __version_less_than(dot_rethink: Dict, version) -> bool:
    dot_version = dot_rethink.get("version")
    if dot_version is None:
        return True
    return version_tuple(dot_version) < version_tuple(version)


def __migrate_older_to_0_2_8(dot_rethink: Dict, data_path: Path):
    """Migrate the database to version 0.2.8.

    Args:
        data_path (str): the path to the database folder.
    """
    v = "0.2.8"
    dot_rethink["version"] = v
    if "settings" not in dot_rethink:
        dot_rethink["settings"] = {
            "language": const.LanguageEnum.EN.value,
            "theme": const.app.AppThemeEnum.LIGHT.value,
            "editorMode": const.app.EditorModeEnum.WYSIWYG.value,
            "editorFontSize": 15,
            "editorCodeTheme": const.app.EditorCodeThemeEnum.GITHUB.value,
            "editorSepRightWidth": 200,
            "editorSideCurrentToolId": "",
        }
    __renew_dot_rethink(dot_rethink, data_path)
    logger.debug(f"Migrate the database to version {v}")


def __read_dot_rethink(data_path: Path) -> Optional[Dict]:
    """Read the .rethink file.

    Args:
        data_path (str): the path to the database folder.

    Returns:
        Dict: the content of the .rethink file.
    """
    dot_rethink = data_path / const.settings.DOT_DATA / ".rethink.json"
    try:
        with open(dot_rethink, "r", encoding="utf-8") as f:
            return json.load(f)
    except FileNotFoundError:
        return None


def __renew_dot_rethink(dot_rethink: Dict, data_path: Path):
    """Rewrite the .rethink file.

    Args:
        dot_rethink (Dict): the content of the .rethink file.
        data_path (str): the path to the database folder.
    """
    with open(data_path / const.settings.DOT_DATA / ".rethink.json", "w", encoding="utf-8") as f:
        json.dump(dot_rethink, f, indent=2, ensure_ascii=False)


def to_latest_version(data_path: Union[str, Path] = "."):
    """Migrate the database to the latest version.

    Args:
        data_path (str): the path to the database folder.
    """
    if not isinstance(data_path, Path):
        data_path = Path(data_path)
    dot_rethink = __read_dot_rethink(data_path)
    if dot_rethink is None:
        logger.debug(".rethink.json is not found. The database is empty.")
        return

    if __version_less_than(dot_rethink, version="0.2.8"):
        __migrate_older_to_0_2_8(dot_rethink, data_path)
        return



from random import randint
from typing import Tuple

from fastapi.responses import StreamingResponse, JSONResponse

from retk import config, const, safety
from retk.controllers import schemas
from retk.controllers.utils import json_exception
from retk.core import account, user, statistic
from retk.models.tps import AuthedUser
from retk.utils import get_token, jwt_encode, jwt_decode


def set_cookie_response(uid: str, req_id: str, status_code: int, access_token: str, refresh_token: str):
    resp = JSONResponse(
        status_code=status_code,
        content={
            "requestId": req_id,
        })
    s = config.get_settings()

    resp.set_cookie(
        key=const.settings.COOKIE_ACCESS_TOKEN,
        value=access_token,
        httponly=True,  # prevent JavaScript from accessing the cookie, XSS
        secure=safety.cookie_secure,  # only send the cookie over HTTPS
        samesite=safety.cookie_samesite,  # prevent CSRF
        expires=s.JWT_ACCESS_EXPIRED_MINS * 60,  # seconds
        domain=safety.cookie_domain,  # prevent CSRF
    )

    if refresh_token != "":
        resp.set_cookie(
            key=const.settings.COOKIE_REFRESH_TOKEN,
            value=refresh_token,
            httponly=True,  # prevent JavaScript from accessing the cookie, XSS
            secure=safety.cookie_secure,  # only send the cookie over HTTPS
            samesite=safety.cookie_samesite,  # prevent CSRF
            expires=s.JWT_REFRESH_EXPIRED_DAYS * 24 * 60 * 60,  # seconds
            domain=safety.cookie_domain,  # prevent CSRF
        )
        resp.set_cookie(
            key=const.settings.COOKIE_REFRESH_TOKEN_ID,
            value=uid,
            httponly=True,  # prevent JavaScript from accessing the cookie, XSS
            secure=safety.cookie_secure,  # only send the cookie over HTTPS
            samesite=safety.cookie_samesite,  # prevent CSRF
            expires=s.JWT_REFRESH_EXPIRED_DAYS * 24 * 60 * 60,  # seconds
            domain=safety.cookie_domain,  # prevent CSRF
        )
    return resp


async def signup(
        req_id: str,
        req: schemas.account.SignupRequest,
) -> JSONResponse:
    if not const.LanguageEnum.is_valid(req.language):
        req.language = const.LanguageEnum.EN.value
    code = account.app_captcha.verify_captcha(token=req.verificationToken, code_str=req.verification)
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )

    new_user, code = await account.manager.signup(
        req.email,
        req.password,
        req.language,
    )
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )

    access_token, refresh_token = get_token(
        uid=new_user["id"],
        language=req.language,
    )
    return set_cookie_response(
        uid=new_user["id"],
        req_id=req_id,
        status_code=201,
        access_token=access_token,
        refresh_token=refresh_token,
    )


async def login(
        req_id: str,
        req: schemas.account.LoginRequest,
) -> JSONResponse:
    # TODO: 后台应记录成功登录用户名和 IP、时间.
    #  当尝试登录 IP 不在历史常登录 IP 地理位置时，应进行多因素二次验证用户身份，防止用户因密码泄漏被窃取账户
    u, code = await user.get_by_email(req.email, disabled=None, exclude_manager=False)
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )
    if u["disabled"]:
        raise json_exception(
            request_id=req_id,
            code=const.CodeEnum.USER_DISABLED,
            language=req.language,
        )

    if not await account.manager.is_right_password(
            email=u["email"],
            hashed=u["hashed"],
            password=req.password,
    ):
        code = const.CodeEnum.ACCOUNT_OR_PASSWORD_ERROR
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )
    access_token, refresh_token = get_token(
        uid=u["id"],
        language=u["settings"]["language"],
    )
    await statistic.add_user_behavior(
        uid=u["id"],
        type_=const.UserBehaviorTypeEnum.LOGIN,
        remark="",
    )
    return set_cookie_response(
        uid=u["id"],
        req_id=req_id,
        status_code=200,
        access_token=access_token,
        refresh_token=refresh_token,
    )


async def auto_login(
        token: str,
        req_id: str,
) -> schemas.user.UserInfoResponse:
    r = schemas.user.UserInfoResponse(
        requestId=req_id,
    )
    if token == "":
        return r
    try:
        payload = jwt_decode(token=token)
    except Exception:  # pylint: disable=broad-except
        return r
    u, code = await user.get(uid=payload["uid"], disabled=False)
    if code != const.CodeEnum.OK:
        return r
    return schemas.user.get_user_info_response_from_u_dict(u, request_id=req_id)


async def logout(
        req_id: str,
        au: AuthedUser,
) -> JSONResponse:
    await statistic.add_user_behavior(
        uid=au.u.id,
        type_=const.UserBehaviorTypeEnum.LOGOUT,
        remark="",
    )

    resp = JSONResponse(
        status_code=200,
        content={
            "requestId": req_id,
        }
    )
    for key in [
        const.settings.COOKIE_ACCESS_TOKEN,
        const.settings.COOKIE_REFRESH_TOKEN,
        const.settings.COOKIE_REFRESH_TOKEN_ID
    ]:
        resp.delete_cookie(
            key=key,
            domain=safety.cookie_domain,
        )

    return resp


async def forget(
        req_id: str,
        req: schemas.account.ForgetPasswordRequest
) -> schemas.RequestIdResponse:
    code = account.email.verify_number(token=req.verificationToken, number_str=req.verification)
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )
    u, code = await user.get_by_email(email=req.email)
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )

    if u is None:
        raise json_exception(
            request_id=req_id,
            code=const.CodeEnum.INVALID_AUTH,
            language=req.language,
        )

    code = await user.reset_password(
        uid=u["id"],
        hashed=account.manager.hash_password(password=req.newPassword, email=req.email)
    )
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=const.LanguageEnum.EN.value,
        )
    return schemas.RequestIdResponse(
        requestId=req_id,
    )


def get_captcha_img():
    token, data = account.app_captcha.generate(length=4, sound=False)
    return StreamingResponse(
        data["img"],
        headers={
            "X-Captcha-Token": token
        },
        media_type="image/png",
    )


def __check_and_send_email(
        email: str,
        token: str,
        code_str: str,
        language: str,
) -> Tuple[str, int, const.CodeEnum]:
    # verify captcha code in image
    code = account.app_captcha.verify_captcha(token=token, code_str=code_str)

    if code != const.CodeEnum.OK:
        return "", 0, code

    numbers = "".join([str(randint(0, 9)) for _ in range(6)])
    expired_min = 10
    code = account.email.email_server.send(
        recipient=email,
        numbers=numbers,
        expire=expired_min,
        language=language,
    )
    return numbers, expired_min, code


async def email_send_code(
        req_id: str,
        req: schemas.account.EmailVerificationRequest
) -> schemas.account.TokenResponse:
    if req.language not in [lang.value for lang in const.LanguageEnum.__members__.values()]:
        req.language = const.LanguageEnum.EN.value

    if not req.userExistOk:
        u, code = await user.get_by_email(email=req.email)
        if u is not None:
            raise json_exception(
                request_id=req_id,
                code=const.CodeEnum.ACCOUNT_EXIST_TRY_FORGET_PASSWORD,
                language=req.language,
            )

    numbers, expired_min, code = __check_and_send_email(
        email=req.email,
        token=req.captchaToken,
        code_str=req.captchaCode,
        language=req.language,
    )
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id=req_id,
            code=code,
            language=req.language,
        )

    token = account.email.encode_number(number=numbers, expired_min=expired_min)
    return schemas.account.TokenResponse(
        requestId=req_id,
        token=token,
    )


async def get_new_access_token(
        au: AuthedUser,
) -> JSONResponse:
    access_token = jwt_encode(
        exp_delta=config.get_settings().ACCESS_TOKEN_EXPIRE_DELTA,
        data={
            "is_access": True,
            "uid": au.u.id,
            "language": au.language,
        },
    )
    return set_cookie_response(
        uid=au.u.id,
        req_id=au.request_id,
        status_code=200,
        access_token=access_token,
        refresh_token="",
    )



import pymongo

from . import mongita_client


def _resolve_client(connection_type, uri):
    """
    :param str connection_type:
    :param str uri:
    :rtype: mongita.MongitaClientDisk|pymongo.MongoClient
    """
    assert connection_type in ('mongita', 'mongodb')
    if connection_type == 'mongita':
        if uri:
            uri = uri.replace('file://', '/')
        return mongita_client.MongitaClientDisk(uri)
    return pymongo.MongoClient(host=uri or 'localhost')


def _resolve_collections(collections):
    """
    Split a list of raw collections into a list of database/collection tuples

    :param list[str] collections:
    :rtype: list[(str, str|None)]
    """
    ret = []
    for raw_collection in collections:
        attr_chain = raw_collection.split('.', 1)
        database = attr_chain[0]
        if len(attr_chain) == 2:
            ret.append((database, attr_chain[1]))
        else:
            ret.append((database, None))
    return ret


def _batch_docs(cursor, cnt=1000):
    """
    Batch a generator of documents into lists of length cnt

    :param generator[dict] cursor:
    :param int cnt:
    :rtype: generator[list[dict]]
    """
    ret = []
    for doc in cursor:
        ret.append(doc)
        if len(ret) == cnt:
            yield ret
            ret = []
    if ret:
        yield ret


def _confirm_loop(msg, logger):
    """
    Confirm in a loop that the user wants to do the thing.
    Returns a tuple of (yes/no, yesall)

    :param str msg:
    :param Logger logger:
    :rtype: (bool, bool)
    """
    while True:
        logger.log("%s (yes/yesall/no)", msg)
        confirm = input()
        if confirm.lower() == 'yesall':
            return True, True
        if confirm.lower() in ('yes', 'y'):
            return True, False
        if confirm.lower() in ('no', 'n'):
            return False, False


def _replace_collection(source, dest, database, collection, force, logger):
    """
    Replace a single collection at destination with the source's collection.
    Returns whether we want to 'force' going forward

    :param MongoClient|MongitaClientDisk source:
    :param MongoClient|MongitaClientDisk dest:
    :param str database:
    :param str collection:
    :param bool force:
    :param Logger logger:
    :rtype: bool
    """
    source_coll = source[database][collection]
    dest_coll = dest[database][collection]
    if force:
        logger.log("Replacing %s.%s (%d documents -> %d documents)...",
                   database, collection, source_coll.count_documents({}),
                   dest_coll.count_documents({}))
    else:
        confirm, force = _confirm_loop("Replace %s.%s? (%d documents -> %d documents)" %
                                       (database, collection,
                                        source_coll.count_documents({}),
                                        dest_coll.count_documents({})),
                                       logger)
        if not confirm:
            return False
    dest[database].drop_collection(collection)
    for doc_batch in _batch_docs(source_coll.find({})):
        dest_coll.insert_many(doc_batch)
    return force


class _Logger():
    def __init__(self, quiet):
        self.quiet = quiet

    def log(self, msg, *args):
        if not self.quiet:
            msg = msg % args
            print("MONGITASYNC: %s" % (msg))


def mongitasync(source_type, destination_type, collections, force=False,
                source_uri=None, destination_uri=None, quiet=False):
    """
    Sync a list of collections from the source to the destination.
    Source/destination can be either 'mongita' or 'mongodb'
    Collections can be formatted as either 'db.coll' or plain 'db'

    :param str source_type: mongita|mongodb
    :param str destination_type: mongita|mongodb
    :param list[str]|str collections:
    :param bool force:
    :param str source_uri:
    :param str destination_uri:
    :param bool quiet:
    """
    source = _resolve_client(source_type, source_uri)
    destination = _resolve_client(destination_type, destination_uri)

    if not collections:
        raise AssertionError("No collections provided")
    if not isinstance(collections, list):
        collections = [collections]

    logger = _Logger(quiet)
    logger.log("Syncing %d databases/collections from %r (%s) to %r (%s):",
               len(collections), source_type, source_uri,
               destination_type, destination_uri)
    for collection in collections:
        logger.log('  ' + collection)

    for database, collection in _resolve_collections(collections):
        if collection:
            force = _replace_collection(source, destination, database, collection,
                                        force, logger)
            continue

        db_collections = list(source[database].list_collection_names())
        if force:
            destination.drop_database(database)
        else:
            confirm, force = _confirm_loop("Drop database %r on %r?" %
                                           (database, destination), logger)
            if confirm:
                destination.drop_database(database)

        for collection in db_collections:
            force = _replace_collection(source, destination, database, collection,
                                        force, logger)



from datetime import datetime
from typing import List, Literal, TypedDict

from bson import ObjectId

CODE_THEME_TYPES = Literal[
    "dracula", "github", "emacs", "tango",
]


class _LastState(TypedDict):
    nodeDisplayMethod: int
    nodeDisplaySortKey: str
    recentSearch: List[str]
    recentCursorSearchSelectedNIds: List[str]  # at search


class _Settings(TypedDict):
    language: str
    theme: Literal["light", "dark"]
    editorMode: Literal["ir", "wysiwyg"]
    editorFontSize: int
    editorCodeTheme: CODE_THEME_TYPES
    editorSepRightWidth: int
    editorSideCurrentToolId: str


class UserMeta(TypedDict):
    _id: ObjectId
    id: str
    source: int
    account: str
    nickname: str
    email: str
    avatar: str
    hashed: str
    disabled: bool
    modifiedAt: datetime
    usedSpace: int
    type: int
    lastState: _LastState
    settings: _Settings



from textwrap import dedent

from .languages import LanguageEnum

DEFAULT_USER = {
    "nickname": "rethink",
    "email": "rethink@rethink.run",
    "avatar": "",
}

NEW_USER_DEFAULT_NODES = {
    LanguageEnum.EN.value: [
        dedent("""\
        # How do I record

        I like to record freely and without any restrictions.
        """),
        dedent("""\
        Welcome to Rethink

        Rethink is a Knowledge Management System. You can take node and manage them here.

        Use @ to link between notes. For example [@How do I record](/n/{}) .

        Rethink also supports markdown syntax, for example:

        # My task list

        - [ ] task 1
        - [ ] task 2
        - [x] task 3

        """),
    ],
    LanguageEnum.ZH.value: [
        dedent("""\
        # 我如何记录

        我喜欢自由自在的记录，不受任何限制。
        """),
        dedent("""\
        欢迎使用 Rethink

        Rethink 是一个知识管理系统，你可以在这里记录你的知识，管理你的记录。

        使用 @，你就可以链接任意的记录，创建联想。比如 [@我如何记录](/n/{}) 。

        Rethink 同样也支持 markdown 语法，让你可以记录更丰富的表达。比如：

        # 我的任务清单

        - [ ] 任务 1
        - [ ] 任务 2
        - [x] 任务 3

        """),
    ]
}



from typing import List, Any

from pydantic import BaseModel, Field

from retk.const import settings


class PluginsResponse(BaseModel):
    class Plugin(BaseModel):
        id: str
        name: str
        version: str
        description: str
        author: str
        iconSrc: str

    requestId: str
    plugins: List[Plugin]


class RenderPluginResponse(BaseModel):
    requestId: str
    html: str


class PluginCallRequest(BaseModel):
    requestId: str = Field(description="request id", max_length=settings.REQUEST_ID_MAX_LENGTH, default="")
    pluginId: str = Field(max_length=settings.PLUGIN_ID_MAX_LENGTH)
    method: str
    data: Any


class PluginCallResponse(BaseModel):
    success: bool
    message: str
    requestId: str
    pluginId: str
    method: str
    data: Any



import shutil
import unittest
from pathlib import Path
from typing import Dict

import retk
from retk import core, const
from retk.models.client import client
from retk.models.tps import convert_user_dict_to_authed_user, AuthedUser
from retk.plugins.base import event_plugin_map
from . import utils


class TestPlugin(retk.Plugin):
    id = "testPlugin"
    name = "TestPlugin"
    version = "0.1.0"
    description = "A demo test plugin."
    author = "morvanzhou"
    icon = "tmp/back.svg"
    template = "<h1>{h}</h1>\n<p>{p}</p>"

    def __init__(self):
        super().__init__()
        self.cache: Dict[str, retk.tps.Node] = {}
        self.bmd = ""
        self.h = ""
        self.p = ""

    def render(self):
        return self.template.format(h=self.h, p=self.p)

    def on_node_added(self, node: retk.tps.Node):
        self.cache[node["id"]] = node

    def before_node_updated(self, uid: str, nid: str, data: Dict[str, str]):
        self.bmd = "before_node_updated:" + data["md"]

    def on_node_updated(self, node: retk.tps.Node, old_node: retk.tps.Node):
        self.cache[node["id"]] = node


svg = """<?xml version="1.0" encoding="utf-8"?>
<svg width="800px" height="800px" viewBox="0 0 1024 1024" xmlns="http://www.w3.org/2000/svg">
    <path fill="#000000" d="M224 480h640a32 32 0 1 1 0 64H224a32 32 0 0 1 0-64z"/>
    <path fill="#000000"
          d="m237.248 512 265.408 265.344a32 32 0 0 1-45.312 45.312l-288-288a32 32 0 0 1 0-45.312l288-288a32
           32 0 1 1 45.312 45.312L237.248 512z"/>
</svg>"""


class DemoCount(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        tmp_dir = Path(__file__).parent / "tmp"
        tmp_dir.mkdir(exist_ok=True)
        with open(tmp_dir / "back.svg", "w") as f:
            f.write(svg)

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")
        shutil.rmtree(Path(__file__).parent / "tmp", ignore_errors=True)

    async def asyncSetUp(self) -> None:
        self.p = TestPlugin()
        retk.add_plugin(self.p)

        await client.init()
        u, _ = await core.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.au = AuthedUser(
            u=convert_user_dict_to_authed_user(u),
            request_id="test",
            language=const.LanguageEnum.EN.value,
        )

    async def asyncTearDown(self) -> None:
        retk.remove_plugin(self.p)
        await client.drop()
        shutil.rmtree(Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "files", ignore_errors=True)
        shutil.rmtree(Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "md", ignore_errors=True)

    def test_creation(self):
        self.assertEqual(self.p.id, "testPlugin")
        self.assertEqual(self.p.name, "TestPlugin")
        self.assertEqual(self.p.version, "0.1.0")
        self.assertEqual(self.p.description, "A demo test plugin.")
        self.assertEqual(self.p.author, "morvanzhou")

    def test_render(self):
        self.p.h = "test h1"
        self.p.p = "test p"
        self.assertEqual(
            "<h1>test h1</h1>\n<p>test p</p>",
            self.p.render()
        )

    def test_set_template_from_file(self):
        self.p.template = "<h2>{h}</h2>\n<p>{p}</p>"
        self.p.h = "test h"
        self.p.p = "test p"
        self.assertEqual(
            "<h2>test h</h2>\n<p>test p</p>",
            self.p.render()
        )

    async def test_event(self):
        self.assertIn(self.p, event_plugin_map["on_node_added"])

        node, code = await core.node.post(
            au=self.au, md="a", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(retk.const.CodeEnum.OK, code)
        self.assertEqual(node, self.p.cache[node["id"]])
        self.assertEqual("a", self.p.cache[node["id"]]["md"])

        self.assertEqual("", self.p.bmd)
        node, _, code = await core.node.update_md(
            au=self.au, nid=node["id"], md="b"
        )
        self.assertEqual(retk.const.CodeEnum.OK, code)
        self.assertEqual(node, self.p.cache[node["id"]])
        self.assertEqual("b", self.p.cache[node["id"]]["md"])
        self.assertEqual("before_node_updated:b", self.p.bmd)

    def test_timing(self):
        with self.assertRaises(ValueError):
            retk.scheduler.timing.every_minute_at(second=-1)
        t = retk.scheduler.timing.every_minute_at(second=1)
        self.assertEqual(1, t.at_second)
        self.assertEqual(0, t.at_minute)



import datetime
from typing import List, Tuple, Sequence, Dict, Any, Literal

from bson.tz_util import utc
from elastic_transport import ObjectApiResponse
from elasticsearch import AsyncElasticsearch, helpers

from retk import const
from retk.config import get_settings
from retk.logger import logger
from retk.models.search_engine.engine import (
    BaseEngine, SearchDoc, SearchResult, RestoreSearchDoc, STOPWORDS,
)
from retk.models.tps import AuthedUser


def datetime2str(dt: datetime.datetime) -> str:
    return dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"


def get_utc_now() -> str:
    now = datetime.datetime.now(tz=utc)
    return datetime2str(now)


class ESSearcher(BaseEngine):
    es: AsyncElasticsearch
    index: str
    properties = {
        "uid": {
            "type": "keyword"
        },
        "title": {
            "type": "text",
            "analyzer": "ik_analyzer",
            "search_analyzer": "ik_smart",
            "fields": {
                "keyword": {
                    "type": "keyword"
                }
            }
        },
        "disabled": {
            "type": "boolean"
        },
        "inTrash": {
            "type": "boolean"
        },
        "body": {
            "type": "text",
            "analyzer": "ik_analyzer",
            "search_analyzer": "ik_smart"
        },
        "modifiedAt": {
            "type": "date",
            "format": "strict_date_optional_time_nanos",
        },
        "createdAt": {
            "type": "date",
            "format": "strict_date_optional_time_nanos",
        }
    }
    analysis = {
        "analyzer": {
            "ik_analyzer": {
                "type": "custom",
                "tokenizer": "ik_max_word",
                "filter": [
                    "lowercase",
                    "english_stop",
                ]
            },
            "search_stop_analyzer": {
                "type": "custom",
                "char_filter": [],
                "tokenizer": "ik_smart",
                "filter": [
                    "lowercase",
                    "english_stop",
                    "cn_stop",
                ]
            },
            "en_analyzer": {
                "type": "custom",
                "tokenizer": "standard",
                "filter": [
                    "lowercase",
                    "english_stop"
                ]
            },

        },
        "filter": {
            "english_stop": {
                "type": "stop",
                "stopwords": "_english_",
            },
            "cn_stop": {
                "type": "stop",
                "stopwords": STOPWORDS,
            }
        }
    }

    def __init__(self):
        super().__init__()
        self.index = ""

    async def connect(self):
        try:
            self.es
        except AttributeError:
            # config the elastic.yml file, disable ssl for local usage, add a user and password
            self.es = AsyncElasticsearch(
                hosts=get_settings().ES_HOSTS.split(","),
                # sniff_on_start=True,
                # sniff_on_node_failure=True,
                # min_delay_between_sniffing=60,
                basic_auth=(get_settings().ES_USER, get_settings().ES_PASSWORD),
                request_timeout=5,
            )
            resp = await self.es.indices.get_alias(index=f"{get_settings().ES_INDEX_ALIAS}-*")
            for index, aliases in resp.body.items():
                if get_settings().ES_INDEX_ALIAS in aliases["aliases"]:
                    self.index = index
                    break

    async def init(self):
        # please install es 8.11.0
        await self.connect()
        info = await self.es.info()
        if info.body["version"]["number"] != "8.11.0":
            raise ValueError("please install es 8.11.0")

        # if no index found, create one
        if self.index == "" or not await self.es.indices.exists(index=self.index):
            # install ik plugin: https://github.com/medcl/elasticsearch-analysis-ik
            self.index = f"{get_settings().ES_INDEX_ALIAS}-1"
            resp = await self.es.indices.create(
                index=self.index,
                aliases={get_settings().ES_INDEX_ALIAS: {}},
                settings={
                    "index": {
                        "number_of_shards": 1,
                        "number_of_replicas": 0,
                        "refresh_interval": "3s",
                    },
                    "analysis": self.analysis,
                },
                mappings={
                    "properties": self.properties,
                }
            )
            if resp.meta.status != 200:
                raise RuntimeError(f"create index failed, resp: {resp}")

        settings = await self.es.indices.get_settings(index=self.index)
        try:
            analysis = settings.body[self.index]["settings"]["index"]["analysis"]
            if analysis != self.analysis:
                await self.reindex()
        except KeyError:
            await self.reindex()

    async def reindex(self):
        await self.connect()
        new_index_num = int(self.index.split("-")[-1]) + 1
        new_index = f"{get_settings().ES_INDEX_ALIAS}-{new_index_num}"

        resp = await self.es.indices.create(
            index=new_index,
            aliases={get_settings().ES_INDEX_ALIAS: {}},
            settings={
                "index": {
                    "number_of_shards": 1,
                    "number_of_replicas": 0,
                    "refresh_interval": "3s",
                },
                "analysis": self.analysis,
            },
            mappings={
                "properties": self.properties,
            }
        )
        if resp.meta.status != 200:
            raise RuntimeError(f"create index for reindexing failed, resp: {resp}")

        resp = await self.es.reindex(
            body={
                "source": {
                    "index": self.index,
                },
                "dest": {
                    "index": new_index,
                }
            }
        )
        if resp.meta.status != 200:
            raise RuntimeError(f"reindex failed, resp: {resp}")
        resp = await self.es.indices.delete(index=self.index)
        if resp.meta.status != 200:
            raise RuntimeError(f"reindex failed, resp: {resp}")

        self.index = new_index
        logger.debug(f"elasticsearch reindex finished, new index: {self.index}")

    async def close(self):
        await self.es.close()

    async def drop(self):
        await self.connect()
        if self.index != "" and await self.es.indices.exists(index=self.index):
            await self.es.indices.delete(index=self.index)
        self.index = ""
        await self.es.close()
        del self.es

    async def add(self, au: AuthedUser, doc: SearchDoc) -> const.CodeEnum:
        now = get_utc_now()
        resp = await self.es.index(
            index=self.index,
            document={
                "uid": au.u.id,
                "title": doc.title,
                "body": doc.body,
                "modifiedAt": now,
                "createdAt": now,
                "disabled": False,
                "inTrash": False,
            },
            id=doc.nid,
        )
        if resp.meta.status != 201:
            logger.error(f"add failed {au.u.id=} {doc.nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def update(self, au: AuthedUser, doc: SearchDoc) -> const.CodeEnum:
        now = get_utc_now()
        resp = await self.es.update(
            index=self.index,
            id=doc.nid,
            body={
                "doc": {
                    "uid": au.u.id,
                    "title": doc.title,
                    "body": doc.body,
                    "modifiedAt": now,
                }
            }
        )
        if resp.meta.status != 200:
            logger.error(f"update failed {au.u.id=} {doc.nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def to_trash(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        resp = await self.es.update(
            index=self.index,
            id=nid,
            body={
                "doc": {
                    "inTrash": True,
                }
            },
            refresh=True,
        )
        if resp.meta.status != 200:
            logger.error(f"to trash failed {au.u.id=} {nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def batch_to_trash(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        resp = await helpers.async_bulk(
            client=self.es,
            actions=[
                {
                    "_op_type": "update",
                    "_index": get_settings().ES_INDEX_ALIAS,
                    "_id": nid,
                    "doc": {
                        "inTrash": True,
                    }
                }
                for nid in nids
            ]
        )
        if resp[0] != len(nids):
            logger.error(f"to trash batch failed, resp: {resp}")
            return const.CodeEnum.OPERATION_FAILED
        await self.refresh()
        return const.CodeEnum.OK

    async def restore_from_trash(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        resp = await self.es.update(
            index=self.index,
            id=nid,
            body={
                "doc": {
                    "inTrash": False,
                }
            },
            refresh=True,
        )
        if resp.meta.status != 200:
            logger.error(f"restore from trash failed {au.u.id=} {nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def restore_batch_from_trash(self, au: AuthedUser, nids: str) -> const.CodeEnum:
        resp = await helpers.async_bulk(
            client=self.es,
            actions=[
                {
                    "_op_type": "update",
                    "_index": get_settings().ES_INDEX_ALIAS,
                    "_id": nid,
                    "doc": {
                        "inTrash": False,
                    }
                }
                for nid in nids
            ]
        )
        if resp[0] != len(nids):
            logger.error(f"restore batch from trash failed, resp: {resp}")
            return const.CodeEnum.OPERATION_FAILED
        await self.refresh()
        return const.CodeEnum.OK

    async def disable(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        resp = await self.es.update(
            index=self.index,
            id=nid,
            body={
                "doc": {
                    "disabled": True,
                }
            },
            refresh=True,
        )
        if resp.meta.status != 200:
            logger.error(f"disable failed {au.u.id=} {nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def enable(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        resp = await self.es.update(
            index=self.index,
            id=nid,
            body={
                "doc": {
                    "disabled": False,
                }
            },
            refresh=True,
        )
        if resp.meta.status != 200:
            logger.error(f"enable failed {au.u.id=} {nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def delete(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        doc = await self.es.get(
            index=self.index,
            id=nid,
        )
        if doc["_source"]["uid"] != au.u.id:
            logger.error(f"node not belong to user {au.u.id=} {nid=}")
            return const.CodeEnum.NODE_NOT_EXIST
        if not doc["_source"]["inTrash"]:
            logger.error(f"doc not in trash, deletion failed {au.u.id=} {nid=}")
            return const.CodeEnum.OPERATION_FAILED

        resp = await self.es.delete(
            index=self.index,
            id=nid,
            refresh=True,
        )
        if resp.meta.status != 201:
            logger.error(f"delete failed {au.u.id=} {nid=}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def add_batch(self, au: AuthedUser, docs: List[SearchDoc]) -> const.CodeEnum:
        actions = []
        now = datetime.datetime.now(tz=utc)
        for doc in docs:
            d = doc.__dict__
            d["inTrash"] = False
            d["disabled"] = False
            d["createdAt"] = datetime2str(now)
            d["modifiedAt"] = d["createdAt"]
            d["uid"] = au.u.id
            nid = d.pop("nid")
            # insert a creation operation
            actions.append({
                "_op_type": "create",
                "_index": get_settings().ES_INDEX_ALIAS,
                "_id": nid,
                "_source": d
            })
            now = now + datetime.timedelta(seconds=0.001)
        return await self._batch_ops(actions, op_type="add", refresh=False)

    async def delete_batch(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        resp = await self.es.delete_by_query(
            index=self.index,
            body={
                "query": {
                    "bool": {
                        "must": [
                            {"ids": {"values": nids}},
                            {
                                "term": {
                                    "uid": au.u.id
                                }
                            },
                            {
                                "term": {
                                    "inTrash": True
                                }
                            }
                        ]
                    }
                }
            },
            refresh=True,
        )

        if resp.meta.status != 200:
            logger.error(f"delete batch failed, resp: {resp}")
            return const.CodeEnum.OPERATION_FAILED
        if resp.body["deleted"] != len(nids):
            logger.error(f"delete batch failed, resp: {resp}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def force_delete_all(self, uid: str) -> const.CodeEnum:
        resp = await self.es.delete_by_query(
            index=self.index,
            body={
                "query": {
                    "bool": {
                        "must": [
                            {
                                "term": {
                                    "uid": uid
                                }
                            },

                        ]
                    }
                }
            },
            refresh=True,
        )
        if resp.meta.status != 200:
            logger.error(f"force delete all failed, resp: {resp}")
            return const.CodeEnum.OPERATION_FAILED
        return const.CodeEnum.OK

    async def update_batch(self, au: AuthedUser, docs: List[SearchDoc]) -> const.CodeEnum:
        actions = []
        now = datetime.datetime.now(tz=utc)
        for doc in docs:
            d = doc.__dict__
            d["modifiedAt"] = datetime2str(now)
            d["uid"] = au.u.id
            nid = d.pop("nid")
            actions.append({
                "_op_type": "update",
                "_index": get_settings().ES_INDEX_ALIAS,
                "_id": nid,
                "doc": d
            })
            now = now + datetime.timedelta(seconds=0.001)
        return await self._batch_ops(actions, op_type="update", refresh=False)

    async def batch_restore_docs(self, au: AuthedUser, docs: List[RestoreSearchDoc]) -> const.CodeEnum:
        actions = []
        for doc in docs:
            d = doc.__dict__
            d["createdAt"] = datetime2str(d["createdAt"])
            d["modifiedAt"] = datetime2str(d["modifiedAt"])
            d["uid"] = au.u.id
            nid = d.pop("nid")
            # insert a creation operation
            actions.append({
                "_op_type": "create",
                "_index": get_settings().ES_INDEX_ALIAS,
                "_id": nid,
                "_source": d
            })
        return await self._batch_ops(actions, "restore", refresh=True)

    async def _search(
            self,
            au: AuthedUser,
            query: str = "",
            sort_key: Literal[
                "createdAt", "modifiedAt", "title", "similarity"
            ] = None,
            reverse: bool = False,
            page: int = 0,
            page_size: int = 10,
            exclude_nids: Sequence[str] = None,
            with_stop_analyzer: bool = False,
    ) -> ObjectApiResponse[Any]:
        if page < 0:
            raise ValueError("page must >= 0")
        if sort_key is None or sort_key in ["", "similarity"]:
            sort = ["_score"]
        elif sort_key == "title":
            sort = [{"title.keyword": {"order": "desc" if reverse else "asc"}}, "_score"]
        elif sort_key in ["createdAt", "modifiedAt"]:
            sort = [
                {sort_key: {
                    "order": "desc" if reverse else "asc",
                    "format": "strict_date_optional_time_nanos"
                }},
                "_score",
            ]
        else:
            raise ValueError(f"sort_key {sort_key} not supported")
        # select uid = au.u.id, disabled = false, inTrash = false
        query_dict: Dict[str, Any] = {
            "bool": {
                "must": [
                    {"constant_score": {"filter": {"term": {"uid": au.u.id}}}},
                    {"constant_score": {"filter": {"term": {"disabled": False}}}},
                    {"constant_score": {"filter": {"term": {"inTrash": False}}}},
                ]
            },
        }
        if query != "":
            q_lower = query.lower()
            analyzer = "search_stop_analyzer" if with_stop_analyzer else "ik_smart"
            # search q_lower on both title and body
            query_dict["bool"]["must"].append({
                "bool": {
                    "should": [
                        {
                            "match": {
                                "title": {
                                    "query": q_lower,
                                    "boost": 2,
                                    "analyzer": analyzer,
                                }
                            }
                        },
                        {
                            "match": {
                                "body": {
                                    "query": q_lower,
                                    "boost": 1,
                                    "analyzer": analyzer,
                                }
                            }
                        }
                    ]
                }
            })
        if exclude_nids is not None and len(exclude_nids) > 0:
            query_dict["bool"]["must_not"] = [{
                "ids": {
                    "values": exclude_nids
                }
            }]

        resp = await self.es.search(
            index=self.index,
            body={
                "query": query_dict,
                "sort": sort,
                "from": page * page_size,
                "size": page_size,
                "highlight": {
                    "pre_tags": [
                        f"<{self.hl_tag_name} class=\"{self.hl_class_name} {self.hl_term_prefix}{i}\">"
                        for i in range(5)
                    ],
                    "post_tags": [f"</{self.hl_tag_name}>" for _ in range(5)],
                    "fields": {
                        "body": {},
                        "title": {},
                    }
                },
            })
        return resp

    async def search(
            self,
            au: AuthedUser,
            query: str = "",
            sort_key: Literal[
                "createdAt", "modifiedAt", "title", "similarity"
            ] = None,
            reverse: bool = False,
            page: int = 0,
            limit: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> Tuple[List[SearchResult], int]:
        resp = await self._search(
            au=au,
            query=query,
            sort_key=sort_key,
            reverse=reverse,
            page=page,
            page_size=limit,
            exclude_nids=exclude_nids,
            with_stop_analyzer=False,
        )
        hits = resp["hits"]["hits"]
        total = resp["hits"]["total"]["value"]
        return [
            SearchResult(
                nid=hit["_id"],
                score=hit["_score"] if hit["_score"] is not None else 0.,
                titleHighlight=self.get_hl(hit, "title", first=True, default=hit["_source"]["title"]),
                bodyHighlights=self.get_hl(hit, "body", first=False, default=hit["_source"]["body"][:60] + "..."),
            ) for hit in hits
        ], total

    async def recommend(
            self,
            au: AuthedUser,
            content: str,
            max_return: int = 10,
            exclude_nids: int = None,
    ) -> List[SearchResult]:
        threshold = 3.
        resp = await self._search(
            au=au,
            query=content,
            sort_key="similarity",
            reverse=False,
            page=0,
            page_size=max_return,
            exclude_nids=exclude_nids,
            with_stop_analyzer=True,
        )
        hits = resp["hits"]["hits"]
        return [
            SearchResult(
                nid=hit["_id"],
                score=hit["_score"],
                titleHighlight=self.get_hl(hit, "title", first=True, default=hit["_source"]["title"]),
                bodyHighlights=self.get_hl(hit, "body", first=False, default=hit["_source"]["body"][:60] + "..."),
            ) for hit in hits if hit["_score"] is not None and hit["_score"] > threshold
        ]

    async def count_all(self) -> int:
        resp = await self.es.count(
            index=self.index,
            body={
                "query": {
                    "match_all": {}
                }
            })
        return resp["count"]

    async def refresh(self):
        await self.es.indices.refresh(index=self.index)

    async def _batch_ops(self, actions: List[dict], op_type: str, refresh: bool) -> const.CodeEnum:
        try:
            resp = await helpers.async_bulk(client=self.es, actions=actions)
        except helpers.BulkIndexError as e:
            logger.error(f"{op_type} batch failed, resp: {e.args[0]}")
            return const.CodeEnum.OPERATION_FAILED
        if resp[0] != len(actions):
            logger.error(f"{op_type} batch failed, resp: {resp}")
            return const.CodeEnum.OPERATION_FAILED
        if refresh:
            await self.refresh()
        return const.CodeEnum.OK

    @staticmethod
    def get_hl(hit: dict, key: str, first: bool, default: str = ""):
        if "highlight" not in hit:
            if first:
                return default
            return []
        if key not in hit["highlight"]:
            if first:
                return default
            return [default]
        if first:
            if len(hit["highlight"][key]) > 0:
                return hit["highlight"][key][0]
            return default
        return hit["highlight"][key]



from enum import IntEnum, Enum, auto, unique


@unique
class NodeDisplayMethodEnum(IntEnum):
    CARD = 0
    LIST = auto()  # 1


@unique
class NodeDisplaySortKeyEnum(str, Enum):
    MODIFIED_AT = "modifiedAt"
    CREATED_AT = "createdAt"
    TITLE = "title"
    SIMILARITY = "similarity"



from retk import const
from retk.models.client import client
from retk.models.tps import AuthedUser


async def added_at_node(
        au: AuthedUser,
        nid: str,
        to_nid: str,
) -> const.CodeEnum:
    # add selected node to recentCursorSearchSelectedNIds
    node_c = {"uid": au.u.id, "id": {"$in": [nid, to_nid]}}

    # try finding node
    ns = await client.coll.nodes.find(node_c).to_list(length=None)
    if len(ns) != 2:
        return const.CodeEnum.NODE_NOT_EXIST

    rns = au.u.last_state.recent_cursor_search_selected_nids
    if to_nid in rns:
        rns.remove(to_nid)
    rns.insert(0, to_nid)
    if len(rns) > 10:
        rns = rns[:10]

    # add to recentCursorSearchSelectedNIds
    res = await client.coll.users.update_one(
        {"id": au.u.id},
        {"$set": {"lastState.recentCursorSearchSelectedNIds": rns}}
    )
    if res.matched_count != 1:
        return const.CodeEnum.OPERATION_FAILED

    return const.CodeEnum.OK


async def put_recent_search(au: AuthedUser, query: str) -> const.CodeEnum:
    rns = au.u.last_state.recent_search
    try:
        rns.remove(query)
    except ValueError:
        pass
    rns.insert(0, query)
    if len(rns) > 20:
        rns = rns[:20]
    _ = await client.coll.users.update_one(
        {"id": au.u.id},
        {"$set": {"lastState.recentSearch": rns}}
    )
    return const.CodeEnum.OK



import io
import zipfile
from dataclasses import dataclass, field
from datetime import datetime
from os.path import normpath
from pathlib import Path
from typing import Tuple, Dict, Optional

from retk import regex, const
from retk.core.files.saver import saver, File
from retk.utils import short_uuid


@dataclass
class UnzipObsidian:
    @dataclass
    class Meta:
        __slots__ = ["filepath", "filename", "title", "file", "size", "duplicate", "created_at"]
        filepath: str
        filename: str
        title: str
        file: bytes
        size: int
        duplicate: bool
        created_at: Optional[datetime]

    md: Dict[str, Meta] = field(default_factory=dict)
    md_full: Dict[str, Meta] = field(default_factory=dict)
    others: Dict[str, Meta] = field(default_factory=dict)
    others_full: Dict[str, Meta] = field(default_factory=dict)


async def __file_ptn_replace_upload(
        uid: str,
        filepath: str,
        filename: str,  # with ext
        md: str,
        others_full: Dict[str, UnzipObsidian.Meta],
        others_name: Dict[str, UnzipObsidian.Meta],
        span: Tuple[int, int]
) -> str:
    """

    Args:
        filepath:
        filename:
        md:
        others_full:
        others_name:
        span:

    Returns:
        md, uploaded_file_size
    """
    try:
        meta = others_name[filename]
    except KeyError:
        try:
            meta = others_full[filepath]
        except KeyError:
            # not inner file
            return md
    file = File(
        filename=meta.filename,
        data=io.BytesIO(meta.file),
    )
    # file type not recognized
    if file.is_unknown_type():
        return md

    url = await saver.save(
        uid=uid,
        file=file
    )

    if url != "":
        img_ptn = "!" if file.type == const.app.FileTypesEnum.IMAGE else ""
        md = f"{md[: span[0]]}{img_ptn}[{filename}]({url}){md[span[1]:]}"
    return md


async def replace_inner_link_and_upload(
        uid: str,
        md: str,
        exist_path2nid: Dict[str, str],
        others_full: Dict[str, UnzipObsidian.Meta],
        others_name: Dict[str, UnzipObsidian.Meta],
) -> str:
    """

    Args:
        uid:
        md:
        exist_path2nid:
        others_full:
        others_name:

    Returns:
        md
    """
    # orig image link to inner image file
    for match in list(regex.MD_IMG.finditer(md))[::-1]:
        span = match.span()
        filename = match.group(1)
        filepath = match.group(2)

        # ![xxx](aa/bb/cc.png)
        md = await __file_ptn_replace_upload(
            uid=uid,
            filepath=filepath,  # full path in zip
            filename=filename,  # filename in md
            md=md,
            others_full=others_full,
            others_name=others_name,
            span=span,
        )

    # files
    for match in list(regex.OBS_INTERNAL_FILE.finditer(md))[::-1]:
        span = match.span()
        filepath = match.group(1)

        # ![[aa/bb/cc.png]]
        # ![[aa.pdf]]
        md = await __file_ptn_replace_upload(
            uid=uid,
            filepath=filepath,
            filename=Path(filepath).name,
            md=md,
            others_full=others_full,
            others_name=others_name,
            span=span,
        )

    # node link
    for match in list(regex.OBS_INTERNAL_LINK.finditer(md))[::-1]:
        span = match.span()
        filename = match.group(1)
        path = f"{filename}.md"
        try:
            nid = exist_path2nid[path]
        except KeyError:
            nid = short_uuid()
            exist_path2nid[path] = nid
        md = f"{md[: span[0]]}[@{filename}](/n/{nid}){md[span[1]:]}"
    return md


def _decode_filename(filepath: str) -> str:
    encodings = ['utf-8', 'gbk', 'cp437']
    for encoding in encodings:
        try:
            return filepath.encode('cp437').decode(encoding)
        except (UnicodeDecodeError, UnicodeEncodeError):
            continue
    return filepath


def unzip_obsidian(zip_bytes: bytes) -> UnzipObsidian:  # noqa: C901
    with zipfile.ZipFile(io.BytesIO(zip_bytes), 'r') as zf:
        filepaths = zf.namelist()
        extracted_files = UnzipObsidian()
        for filepath in filepaths:
            info = zf.getinfo(filepath)
            if info.is_dir():
                continue
            try:
                _filepath = _decode_filename(filepath)
            except UnicodeDecodeError:
                _filepath = filepath
            fp = Path(_filepath.split("/", 1)[-1])
            # filter out system files
            has_invalid_suffix = False
            for p in fp.parts:
                if p.startswith(".") or fp.name.startswith("__MACOSX"):
                    has_invalid_suffix = True
                    break
            if has_invalid_suffix:
                continue

            # Check for directory traversal
            norm_path = normpath(fp)
            if norm_path.startswith('../') or norm_path.startswith('..\\'):
                continue  # or raise an exception
            try:
                title = regex.MD_FILE_EXT.findall(fp.name)[0]
            except IndexError:
                title = fp.name

            meta = UnzipObsidian.Meta(
                filepath=norm_path,
                filename=fp.name,
                title=title,
                file=zf.read(filepath),
                size=info.file_size,
                duplicate=False,
                created_at=datetime(*info.date_time),
            )

            if fp.suffix.lower() == ".md":
                entry = extracted_files.md
                entry_full = extracted_files.md_full
            else:
                entry = extracted_files.others
                entry_full = extracted_files.others_full

            if fp.name in entry:
                meta.duplicate = True
            else:
                entry[fp.name] = meta
            entry_full[str(fp)] = meta

    return extracted_files



from enum import IntEnum, unique


@unique
class UserSourceEnum(IntEnum):
    TEST = 0
    EMAIL = 1
    PHONE = 2
    GOOGLE = 3
    FACEBOOK = 4
    WECHAT = 5
    GITHUB = 6
    LOCAL = 7

    def __str__(self):
        return self.value



import smtplib

from retk import config


def send(
        recipients: list,
        subject: str,
) -> str:
    server = None
    n = 3
    for _ in range(n):
        try:
            server = smtplib.SMTP('smtp.office365.com', 587)
        except smtplib.SMTPServerDisconnected:
            pass
    if server is None:
        return f"try send email task {n} times, but failed, SMTPServerDisconnected"

    server.ehlo()
    server.starttls()
    settings = config.get_settings()
    try:
        server.login(settings.RETHINK_EMAIL, password=settings.RETHINK_EMAIL_PASSWORD)
    except smtplib.SMTPAuthenticationError as e:
        server.quit()
        return f"SMTPAuthenticationError: {e}"
    try:
        server.sendmail(settings.RETHINK_EMAIL, recipients, subject)
    except smtplib.SMTPRecipientsRefused as e:
        server.quit()
        return f"SMTPRecipientsRefused: {e}"
    server.quit()
    return "done"



from typing import List

from fastapi import UploadFile

from retk import const, core
from retk.controllers import schemas
from retk.controllers.utils import (
    AuthedUser,
    is_allowed_mime_type,
    maybe_raise_json_exception,
)
from retk.utils import datetime2str


async def upload_obsidian_files(
        au: AuthedUser,
        files: List[UploadFile]
) -> schemas.RequestIdResponse:
    code = await core.files.upload_obsidian(au=au, zipped_files=files)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def upload_text_files(
        au: AuthedUser,
        files: List[UploadFile]
) -> schemas.RequestIdResponse:
    code = await core.files.upload_text(au=au, files=files)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def get_upload_process(
        au: AuthedUser,
) -> schemas.files.FileUploadProcessResponse:
    doc = await core.files.get_upload_process(uid=au.u.id)
    if doc is None:
        code = const.CodeEnum.OK
        return schemas.files.FileUploadProcessResponse(
            code=code.value,
            msg=const.get_msg_by_code(code, au.language),
            requestId=au.request_id,
            process=0.,
            type="",
            startAt="",
            running=False,
        )

    return schemas.files.FileUploadProcessResponse(
        code=doc["code"],
        msg=doc["msg"],
        requestId=au.request_id,
        process=doc["process"],
        type=doc["type"],
        startAt=datetime2str(doc["startAt"]),
        running=doc["running"],
    )


async def upload_file_vditor(
        au: AuthedUser,
        file: UploadFile,
) -> schemas.files.VditorFilesResponse:
    res = await core.files.vditor_upload(au=au, files=[file])
    return schemas.files.VditorFilesResponse(
        msg=const.get_msg_by_code(res["code"], au.language),
        code=res["code"].value,
        requestId=au.request_id,
        data=schemas.files.VditorFilesResponse.Data(
            errFiles=res["errFiles"],
            succMap=res["succMap"],
        ),
    )


async def fetch_image_vditor(
        au: AuthedUser,
        req: schemas.files.ImageVditorFetchRequest,
) -> schemas.files.VditorImagesResponse:
    if is_allowed_mime_type(req.url, ["image/svg+xml", "image/png", "image/jpeg", "image/gif"]):
        code = const.CodeEnum.OK
        return schemas.files.VditorImagesResponse(
            msg=const.get_msg_by_code(code, au.language),
            code=code.value,
            requestId=au.request_id,
            data=schemas.files.VditorImagesResponse.Data(
                originalURL=req.url,
                url=req.url,
            ),
        )

    if len(req.url) <= 2048:
        new_url, code = await core.files.fetch_image_vditor(au=au, url=req.url)
    else:
        new_url, code = "", const.CodeEnum.URL_TOO_LONG
    maybe_raise_json_exception(au=au, code=code)
    return schemas.files.VditorImagesResponse(
        msg=const.get_msg_by_code(code, au.language),
        code=code.value,
        requestId=au.request_id,
        data=schemas.files.VditorImagesResponse.Data(
            originalURL=req.url,
            url=new_url,
        ),
    )



from pydantic import BaseModel


class SystemNoticeResponse(BaseModel):
    class Notice(BaseModel):
        id: str
        title: str
        html: str
        publishAt: str

    requestId: str
    notice: Notice



"""SSO login base dependency
"""
# pylint: disable=too-few-public-methods

import json
import sys
import warnings
from types import TracebackType
from typing import Any, Dict, List, Optional, Type

import httpx
import pydantic
from oauthlib.oauth2 import WebApplicationClient
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import RedirectResponse

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict  # pragma: no cover

DiscoveryDocument = TypedDict(
    "DiscoveryDocument", {"authorization_endpoint": str, "token_endpoint": str, "userinfo_endpoint": str}
)


class UnsetStateWarning(UserWarning):
    """Warning about unset state parameter"""


class ReusedOauthClientWarning(UserWarning):
    """Warning about reused oauth client instance"""


class SSOLoginError(HTTPException):
    """Raised when any login-related error ocurrs
    (such as when user is not verified or if there was an attempt for fake login)
    """


class OpenID(pydantic.BaseModel):  # pylint: disable=no-member
    """Class (schema) to represent information got from sso provider in a common form."""

    id: Optional[str] = None
    email: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    display_name: Optional[str] = None
    picture: Optional[str] = None
    provider: Optional[str] = None


# pylint: disable=too-many-instance-attributes
class SSOBase:
    """Base class (mixin) for all SSO providers"""

    provider: str = NotImplemented
    client_id: str = NotImplemented
    client_secret: str = NotImplemented
    redirect_uri: Optional[str] = NotImplemented
    scope: List[str] = NotImplemented
    additional_headers: Optional[Dict[str, Any]] = None

    def __init__(
            self,
            client_id: str,
            client_secret: str,
            redirect_uri: Optional[str] = None,
            allow_insecure_http: bool = False,
            use_state: bool = False,
            scope: Optional[List[str]] = None,
    ):
        # pylint: disable=too-many-arguments
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.allow_insecure_http = allow_insecure_http
        self._oauth_client: Optional[WebApplicationClient] = None
        if use_state:
            warnings.warn(
                (
                    "Argument 'use_state' of SSOBase's constructor is deprecated and will be removed in "
                    "future releases. Use 'state' argument of individual methods instead."
                ),
                DeprecationWarning,
            )
        self.scope = scope or self.scope
        self._refresh_token: Optional[str] = None
        self._state: Optional[str] = None

    @property
    def state(self) -> Optional[str]:
        """Gets state as it was returned from the server"""
        if self._state is None:
            warnings.warn(
                "'state' parameter is unset. This means the server either "
                "didn't return state (was this expected?) or 'verify_and_process' hasn't been called yet.",
                UnsetStateWarning,
            )
        return self._state

    @property
    def oauth_client(self) -> WebApplicationClient:
        """OAuth Client to help us generate requests and parse responses"""
        if self.client_id == NotImplemented:
            raise NotImplementedError(f"Provider {self.provider} not supported")  # pragma: no cover
        if self._oauth_client is None:
            self._oauth_client = WebApplicationClient(self.client_id)
        return self._oauth_client

    @property
    def access_token(self) -> Optional[str]:
        """Access token from token endpoint"""
        return self.oauth_client.access_token

    @property
    def refresh_token(self) -> Optional[str]:
        """Get refresh token (if returned from provider)"""
        return self._refresh_token or self.oauth_client.refresh_token

    @classmethod
    async def openid_from_response(cls, response: dict) -> OpenID:
        """Return {OpenID} object from provider's user info endpoint response"""
        raise NotImplementedError(f"Provider {cls.provider} not supported")

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get discovery document containing handy urls"""
        raise NotImplementedError(f"Provider {self.provider} not supported")

    @property
    async def authorization_endpoint(self) -> Optional[str]:
        """Return `authorization_endpoint` from discovery document"""
        discovery = await self.get_discovery_document()
        return discovery.get("authorization_endpoint")

    @property
    async def token_endpoint(self) -> Optional[str]:
        """Return `token_endpoint` from discovery document"""
        discovery = await self.get_discovery_document()
        return discovery.get("token_endpoint")

    @property
    async def userinfo_endpoint(self) -> Optional[str]:
        """Return `userinfo_endpoint` from discovery document"""
        discovery = await self.get_discovery_document()
        return discovery.get("userinfo_endpoint")

    async def get_login_url(
            self,
            *,
            redirect_uri: Optional[str] = None,
            params: Optional[Dict[str, Any]] = None,
            state: Optional[str] = None,
    ) -> str:
        """Return prepared login url. This is low-level, see {get_login_redirect} instead."""
        params = params or {}
        redirect_uri = redirect_uri or self.redirect_uri
        if redirect_uri is None:
            raise ValueError("redirect_uri must be provided, either at construction or request time")
        request_uri = self.oauth_client.prepare_request_uri(
            await self.authorization_endpoint, redirect_uri=redirect_uri, state=state, scope=self.scope, **params
        )
        return request_uri

    async def get_login_redirect(
            self,
            *,
            redirect_uri: Optional[str] = None,
            params: Optional[Dict[str, Any]] = None,
            state: Optional[str] = None,
    ) -> RedirectResponse:
        """Return redirect response by Stalette to login page of Oauth SSO provider

        Arguments:
            redirect_uri {Optional[str]} -- Override redirect_uri specified on this instance (default: None)
            params {Optional[Dict[str, Any]]} -- Add additional query parameters to the login request.
            state {Optional[str]} -- Add state parameter. This is useful if you want
                                    the server to return something specific back to you.

        Returns:
            RedirectResponse -- Starlette response (may directly be returned from FastAPI)
        """
        login_uri = await self.get_login_url(redirect_uri=redirect_uri, params=params, state=state)
        response = RedirectResponse(login_uri, 303)
        return response

    async def verify_and_process(
            self,
            request: Request,
            *,
            params: Optional[Dict[str, Any]] = None,
            headers: Optional[Dict[str, Any]] = None,
            redirect_uri: Optional[str] = None,
    ) -> Optional[OpenID]:
        """Get FastAPI (Starlette) Request object and process login.
        This handler should be used for your /callback path.

        Arguments:
            request {Request} -- FastAPI request object (or Starlette)
            params {Optional[Dict[str, Any]]} -- Optional additional query parameters to pass to the provider

        Returns:
            Optional[OpenID] -- OpenID if the login was successfull
        """
        headers = headers or {}
        code = request.query_params.get("code")
        if code is None:
            raise SSOLoginError(400, "'code' parameter was not found in callback request")
        self._state = request.query_params.get("state")
        return await self.process_login(
            code, request, params=params, additional_headers=headers, redirect_uri=redirect_uri
        )

    def __enter__(self) -> "SSOBase":
        self._oauth_client = None
        self._refresh_token = None
        return self

    def __exit__(
            self,
            _exc_type: Optional[Type[BaseException]],
            _exc_val: Optional[BaseException],
            _exc_tb: Optional[TracebackType],
    ) -> None:
        return None

    async def process_login(
            self,
            code: str,
            request: Request,
            *,
            params: Optional[Dict[str, Any]] = None,
            additional_headers: Optional[Dict[str, Any]] = None,
            redirect_uri: Optional[str] = None,
    ) -> Optional[OpenID]:
        """This method should be called from callback endpoint to verify the user and request user info endpoint.
        This is low level, you should use {verify_and_process} instead.

        Arguments:
            params {Optional[Dict[str, Any]]} -- Optional additional query parameters to pass to the provider
            additional_headers {Optional[Dict[str, Any]]} -- Optional additional headers to be added to all requests
        """
        # pylint: disable=too-many-locals
        if self._oauth_client is not None:  # pragma: no cover
            self._oauth_client = None
            self._refresh_token = None
            warnings.warn(
                (
                    "Reusing the SSO object is not safe and caused a security issue in previous versions."
                    "To make sure you don't see this warning, please use the SSO object as a context manager."
                ),
                ReusedOauthClientWarning,
            )
        params = params or {}
        additional_headers = additional_headers or {}
        additional_headers.update(self.additional_headers or {})
        url = request.url
        scheme = url.scheme
        if not self.allow_insecure_http and scheme != "https":
            current_url = str(url).replace("http://", "https://")
            scheme = "https"
        else:
            current_url = str(url)
        current_path = f"{scheme}://{url.netloc}{url.path}"

        token_url, headers, body = self.oauth_client.prepare_token_request(
            await self.token_endpoint,
            authorization_response=current_url,
            redirect_url=redirect_uri or self.redirect_uri or current_path,
            code=code,
            **params,
        )  # type: ignore

        if token_url is None:  # pragma: no cover
            return None

        headers.update(additional_headers)

        auth = httpx.BasicAuth(self.client_id, self.client_secret)
        async with httpx.AsyncClient() as session:
            response = await session.post(token_url, headers=headers, content=body, auth=auth)
            content = response.json()
            self._refresh_token = content.get("refresh_token")
            self.oauth_client.parse_request_body_response(json.dumps(content))

            uri, headers, _ = self.oauth_client.add_token(await self.userinfo_endpoint)
            response = await session.get(uri, headers=headers)
            content = response.json()

        return await self.openid_from_response(content)



import unittest
from io import BytesIO

import bcrypt

from retk import const, regex
from retk.core import account, user
from retk.models.client import client
from retk.models.tps import convert_user_dict_to_authed_user
from retk.utils import jwt_decode
from . import utils


class AccountTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        utils.set_env(".env.test.local")
        cls.salt = bcrypt.gensalt(rounds=5)

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    async def asyncSetUp(self) -> None:
        await client.drop()
        await client.init()

    def test_verify(self):
        password = "123abc"
        email = "rethink@rethink.run"
        bpw = account.manager._base_password(password=password, email=email)
        token_str = bcrypt.hashpw(bpw, self.salt).decode("utf-8")
        self.assertEqual(60, len(token_str))
        hashed = account.manager._base_password(password=password, email=email)
        match = bcrypt.checkpw(hashed, token_str.encode("utf-8"))
        self.assertTrue(match)

    def test_hash(self):
        pwd = "123abc&&rethink@rethink.run"
        pwd_bt = pwd.encode("utf-8")
        hashed = bcrypt.hashpw(pwd_bt, self.salt)
        # print(hashed)
        match = bcrypt.checkpw(pwd_bt, hashed)
        self.assertTrue(match)

    async def test_one_user(self):
        u, code = await account.manager.signup("a@q.com", "rethink", const.LanguageEnum.EN.value)
        self.assertEqual(const.CodeEnum.ONE_USER_MODE, code)
        self.assertIsNone(u)

    async def test_verify_user(self):
        u, err = await user.get_by_email("rethink@rethink.run")
        self.assertEqual(const.CodeEnum.OK, err)
        self.assertIsNotNone(u)
        au = convert_user_dict_to_authed_user(u)
        ok = await account.manager.is_right_password(
            email=au.email,
            hashed=au.hashed,
            password="rethink"
        )
        self.assertTrue(ok)

    def test_valid_password(self):
        for t, b in [
            ("rethink", False),
            ("123334", False),
            ("ret123", True),
            ("re123", False),
            ("rethink123", True),
            ("sd@d", False),
            ("sd@d123", True),
            ("1s", False),
            ("", False)
        ]:
            if b:
                self.assertIsNotNone(regex.VALID_PASSWORD.match(t), msg=t)
            else:
                self.assertIsNone(regex.VALID_PASSWORD.match(t), msg=t)

    def test_salt(self):
        npw = "12345"
        self.assertLessEqual(len(npw), const.settings.PASSWORD_MAX_LENGTH)
        bpw = account.manager._base_password(password=npw, email="rethink@rethink.run")
        salt = bcrypt.gensalt()
        hpw = bcrypt.hashpw(bpw, salt).decode("utf-8")
        self.assertEqual(salt.decode("utf-8"), hpw[:len(salt)])
        self.assertNotEqual(bpw.decode("utf-8"), hpw[len(salt):])

    def test_verification_img(self):
        token, data = account.app_captcha.generate()
        self.assertIsInstance(token, str)
        self.assertIn("img", data)
        self.assertIsInstance(data["img"], BytesIO)
        self.assertNotIn("sound", data)

    def test_verify_img(self):
        token, _ = account.app_captcha.generate()
        decoded = jwt_decode(token)
        code = account.app_captcha.verify_captcha(token=token, code_str=decoded["code"])
        self.assertEqual(const.CodeEnum.OK, code)

        code = account.app_captcha.verify_captcha(token=token, code_str="1234")
        self.assertEqual(const.CodeEnum.CAPTCHA_ERROR, code)



from fastapi import APIRouter

from retk.controllers import schemas, statistic
from retk.routes import utils

router = APIRouter(
    prefix="/api/statistic",
    tags=["statistic"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/user-behavior",
    status_code=201,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def add_user_behavior(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.statistic.UserBehaviorRequest,
) -> schemas.RequestIdResponse:
    return await statistic.add_user_behavior(
        au=au,
        req=req,
    )



from .common import support_alert
from .errors import MongitaNotImplementedError


class CommandCursor():
    UNIMPLEMENTED = ['address', 'alive', 'batch_size', 'cursor_id', 'session']

    def __init__(self, _generator):
        self._generator = _generator
        self._cursor = None

    def __getattr__(self, attr):
        if attr in self.UNIMPLEMENTED:
            raise MongitaNotImplementedError.create("CommandCursor", attr)
        raise AttributeError()

    async def __anext__(self):
        return next(await self._gen())

    async def _gen(self):
        """
        This exists so that we can maintain our position in the cursor and
        to not execute until we start requesting items
        """
        if self._cursor:
            return self._cursor
        self._cursor = await self._generator()
        return self._cursor

    @support_alert
    def next(self):
        """
        Returns the next item in the CommandCursor. Raises StopIteration if there
        are no more items.
        """
        return next(self._gen())

    @support_alert
    def close(self):
        """
        Close this cursor to free the memory
        """

        self._cursor = iter(())



from fastapi import APIRouter
from fastapi.params import Path

from retk.controllers import schemas, notice
from retk.routes import utils

router = APIRouter(
    prefix="/api/notices",
    tags=["notices"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/system/{notice_id}",
    status_code=200,
    response_model=schemas.notice.SystemNoticeResponse,
)
@utils.measure_time_spend
async def get_system_notice(
        au: utils.ANNOTATED_AUTHED_USER,
        notice_id: str = Path(max_length=24, title="The ID of the notice"),
) -> schemas.notice.SystemNoticeResponse:
    return await notice.get_system_notice(
        au=au,
        notice_id=notice_id,
    )



import unittest
from textwrap import dedent
from unittest.mock import patch

import httpx

from retk import const, config, utils


class UtilsTest(unittest.TestCase):
    def setUp(self) -> None:
        config.get_settings.cache_clear()

    def tearDown(self) -> None:
        config.get_settings.cache_clear()

    def test_short_uuid(self):
        for _ in range(10):
            uid = utils.short_uuid()
            self.assertEqual(24, len(uid), msg=uid)

    def test_md2txt(self):
        text = utils.md2txt(md="# 123\n## 456\n### 789\n")
        self.assertEqual("123\n456\n789", text)

    def test_change_link_title(self):
        md = dedent("""\
            # 123
            ddd qwd [@我是谁](/n/sqaaq) 345
            [@ssw](/n/weee)
            """)

        new_md = utils.change_link_title(md, nid="sqaaq", new_title="哇塞")
        self.assertEqual(dedent("""\
            # 123
            ddd qwd [@哇塞](/n/sqaaq) 345
            [@ssw](/n/weee)
            """), new_md)

        new_md = utils.change_link_title(md, nid="weee", new_title="哇塞")
        self.assertEqual(dedent("""\
            # 123
            ddd qwd [@我是谁](/n/sqaaq) 345
            [@哇塞](/n/weee)
            """), new_md)

        md = dedent("""\
            # 123
            ddd qwd [@我是谁](/n/sqaaq) 345
            [@我是谁](/n/sqaaq)
            """)
        new_md = utils.change_link_title(md, nid="sqaaq", new_title="哇塞")
        self.assertEqual(dedent("""\
            # 123
            ddd qwd [@哇塞](/n/sqaaq) 345
            [@哇塞](/n/sqaaq)
            """), new_md)

        md = dedent("""\
            # 123
            ddd qwd 345
            [@我](/n/1) [@我](/n/2)
            """)
        new_md = utils.change_link_title(md, nid="2", new_title="哇塞")
        self.assertEqual(dedent("""\
            # 123
            ddd qwd 345
            [@我](/n/1) [@哇塞](/n/2)
            """), new_md)

    def test_contain_only_link(self):
        for md, res in [
            ("http://rethink.run", "http://rethink.run"),
            ("[123](/n/123)", ""),
            ("few", ""),
            ("c https://rethink.run", ""),
            (" https://rethink.run  ", "https://rethink.run"),
            ("https://rethink.run  ", "https://rethink.run"),
            ("https://rethink.run", "https://rethink.run"),
            ("https://rethink.run wwq", ""),

        ]:
            self.assertEqual(res, utils.contain_only_http_link(md))

    def test_strip_html_tags(self):
        for html, res in [
            ("", ""),
            ("a", "a"),
            ("<a>", ""),
            ("<a>123</a>", "123"),
            ("<a>123</a>456", "123456"),
            ("<a>123</a>456<b>789</b>", "123456789"),
            ("<a>123</a>456<b>789</b><c>000</c>", "123456789000"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d>", "123456789000111"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d><e>222</e>", "123456789000111222"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d><e>222</e><f>333</f>", "123456789000111222333"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d><e>222</e><f>333</f><g>444</g>",
             "123456789000111222333444"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d><e>222</e><f>333</f><g>444</g><h>555</h>",
             "123456789000111222333444555"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d><e>222</e><f>333</f><g>444</g><h>555</h><i>666</i>",
             "123456789000111222333444555666"),
            ("<a>123</a>456<b>789</b><c>000</c><d>111</d><e>222</e><f>333</f><g>444</g><h>555</h><i>666</i><j>777</j>",
             "123456789000111222333444555666777"),
            ("d" * 1000000, "d" * 1000)
        ]:
            self.assertEqual(res, utils.strip_html_tags(html))


class TestAsync(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        config.get_settings.cache_clear()

    @classmethod
    def tearDownClass(cls) -> None:
        config.get_settings.cache_clear()

    # @unittest.skip("skip outer connection test")
    @patch("retk.utils.httpx.AsyncClient.get")
    @patch("retk.config.get_settings")
    async def test_get_title_description_from_link(self, mock_get_settings, mock_get, ):
        s = config.Settings
        s.COS_BUCKET_NAME = "rethink-dev-1258395282"
        s.COS_REGION = "ap-hongkong"
        s.DB_HOST = "127.0.0.1"
        mock_get_settings.return_value = s
        for url, content, res in [
            (
                    "https://github.com/MorvanZhou/rethink",
                    "<title>MorvanZhou/rethink: Rethink: a note taking web app</title>"
                    """<meta name="description" content="Rethink: a note taking web app. Contribute to
                     MorvanZhou/rethink development by creating an account on GitHub.">""",
                    True
            ),
            (
                    "https://zhuanlan.zhihu.com/p/610939462?utm_id=0",
                    """<head>
                    <meta charSet="utf-8"/>
                    <title data-rh="true">python的httpx库如何使用 - 知乎</title>
                    <meta data-rh="true" name="description" content="httpx是一个基于Python的异步HTTP客户端库，
                    可以用于发送HTTP请求和接收HTTP响应。以下是一些httpx库的基本使用方法：
                     发送HTTP GET请求import httpx async with httpx.AsyncClient() as client: response = await…"/>""",
                    True
            ),
            (
                    "https://waqwe12f2f2fa.fffffffff",
                    "",
                    False
            ),
            (
                    "https://baidu.com",
                    """<title>百度一下，你就知道</title>
                    <meta name="description" content="全球领先的中文搜索引擎、
                    致力于让网民更便捷地获取信息，找到所求。百度超过千亿的中文网页数据库，可以瞬间找到相关的搜索结果。">""",
                    True
            ),
            (
                    "https://rethink.run",
                    """<meta content="Rethink" name="title"><title>rethink</title>
                    <meta content="Rethink: think differently" name="description">""",
                    True
            ),
            (
                    "https://baidu.com/wqwqqqqq",
                    "",
                    False
            ),
            (
                    "https://mp.weixin.qq.com/s/jbB0GXbjHpFR8m1-6TSASw",
                    """<title></title><meta name="description" content="" />""",
                    False),
            (
                    "http://127.0.0.1",
                    "",
                    False,
            ),
            (
                    "http://127.0.2.1",
                    "",
                    False,
            ),
            (
                    "http://9.0.0.1.xip.io/?id=1",
                    "",
                    False,
            ),
        ]:
            if res:
                mock_get.return_value = httpx.Response(
                    status_code=200,
                    content=content.encode("utf-8"),
                )
            else:
                mock_get.return_value = httpx.Response(
                    status_code=404,
                    content=content.encode("utf-8"),
                )
            title, desc = await utils.get_title_description_from_link(
                url, language=const.LanguageEnum.EN.value)
            if res:
                self.assertNotEqual("No title found", title, msg=f"{url} {title}")
                self.assertNotEqual("No description found", desc, msg=f"{url} {desc}")
            else:
                self.assertEqual("No title found", title, msg=f"{url} {title}")
                self.assertEqual("No description found", desc, msg=f"{url} {desc}")

    def test_mask_email(self):
        for email, res in [
            ("", ""),
            ("a", "a"),
            ("a@b", "a**@b"),
            ("ab@b", "a**b@b"),
            ("abc@b", "ab**c@b"),
            ("abcd@b", "ab**d@b"),
            ("abcdef@b", "ab**f@b"),
        ]:
            self.assertEqual(res, utils.mask_email(email))

    @patch(
        "retk.config.get_settings",
    )
    def test_ssrf(self, mock_get_settings):
        for pre, url in [
            (
                    "rethink-product-1258395282",
                    "https://rethink-product-1258395282.cos.ap-hongkong.myqcloud.com/userData/Rro/as.png",
            ),
            (
                    "rethink-dev-1258395282",
                    "https://rethink-dev-1258395282.cos.ap-hongkong.myqcloud.com/userData/qwqd/qwww.png",
            ),
        ]:
            s = config.Settings
            s.COS_BUCKET_NAME = pre
            s.COS_REGION = "ap-hongkong"
            s.DB_HOST = "127.0.0.1"
            mock_get_settings.return_value = s
            self.assertTrue(utils.ssrf_check(url), msg=url)



import shutil
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Tuple

from bson.tz_util import utc
from pydantic_settings import BaseSettings

try:
    from qcloud_cos import CosConfig, CosServiceError, CosS3Client
except ImportError:
    pass

from retk import config, const
from retk.logger import logger
from retk.models import tps
from retk.models.client import client


async def storage_md(node: tps.Node, keep_hist: bool) -> const.CodeEnum:  # noqa: C901
    nid = node["id"]
    md = node["md"]
    hist = node.get("history", [])
    date_format = "%Y-%m-%d %H:%M:%S.%fZ"

    if config.is_local_db():
        # save to local
        md_path = Path(
            config.get_settings().RETHINK_LOCAL_STORAGE_PATH
        ) / const.settings.DOT_DATA / "md" / (nid + ".md")
        md_path.write_text(md, encoding="utf-8")

    if not keep_hist:
        return const.CodeEnum.OK

    # ignore if the time difference is less than n minutes
    if len(hist) != 0:
        last_hist = hist[0]
        time = datetime.strptime(last_hist, date_format).replace(tzinfo=None)
        if (datetime.now(tz=utc).replace(tzinfo=None) - time) < timedelta(
                seconds=config.get_settings().MD_BACKUP_INTERVAL
        ):
            return const.CodeEnum.OK

    this_hist = node["modifiedAt"].strftime(date_format)
    hist.insert(0, this_hist)

    if config.is_local_db():
        md_dir = __get_md_hist_dir(nid)
        md_dir.mkdir(parents=True, exist_ok=True)
        filename = __windows_safe_path(this_hist)
        with open(md_dir / f"{filename}.md", "w", encoding="utf-8") as f:
            f.write(md)
    else:
        code = __save_md_to_cos(node["uid"], nid, this_hist, md)
        if code != const.CodeEnum.OK:
            return code

    if len(hist) > const.settings.MAX_MD_BACKUP_VERSIONS:
        drop = hist[-const.settings.MAX_MD_BACKUP_VERSIONS:]
        for h in drop:
            if config.is_local_db():
                md_dir = __get_md_hist_dir(nid)
                filename = __windows_safe_path(h)
                (md_dir / f"{filename}.md").unlink()
            else:
                __remove_md_from_cos(node["uid"], nid, h)
        hist = hist[:-const.settings.MAX_MD_BACKUP_VERSIONS]

    res = await client.coll.nodes.update_one(
        {"id": nid},
        {"$set": {"history": hist}}
    )
    if res.modified_count != 1:
        logger.error(f"failed to update node history: {nid}")
        return const.CodeEnum.OPERATION_FAILED
    return const.CodeEnum.OK


def delete_node_md(uid: str, nids: List[str]):
    if config.is_local_db():
        for nid in nids:
            dir_ = __get_md_hist_dir(nid)
            # remove dir even if it's not empty
            shutil.rmtree(dir_, ignore_errors=True)
            # remove current node md
            md_path = Path(
                config.get_settings().RETHINK_LOCAL_STORAGE_PATH
            ) / const.settings.DOT_DATA / "md" / (nid + ".md")
            md_path.unlink(missing_ok=True)
    else:
        for nid in nids:
            __remove_md_all_versions_from_cos(uid, nid)


def get_md(uid: str, nid: str, version: str) -> Tuple[str, const.CodeEnum]:
    if config.is_local_db():
        md_dir = __get_md_hist_dir(nid)
        filename = __windows_safe_path(version)
        with open(md_dir / f"{filename}.md", "r", encoding="utf-8") as f:
            md = f.read()
    else:
        md, code = __get_md_from_cos(uid, nid, version)
        if code != const.CodeEnum.OK:
            return "", code
    return md, const.CodeEnum.OK


def __windows_safe_path(filename: str) -> str:
    return filename.replace(":", "_").replace("\\", "_").replace("/", "_")


def __get_md_hist_dir(nid: str = None) -> Path:
    p = Path(config.get_settings().RETHINK_LOCAL_STORAGE_PATH) / const.settings.DOT_DATA / "md" / "hist"
    if nid is not None:
        p = p / nid
    return p


def __get_md_from_cos(uid: str, nid: str, version: str) -> Tuple[str, const.CodeEnum]:
    settings = config.get_settings()
    cos_client, key = __get_client_and_key(settings, uid, nid, version)

    try:
        res = cos_client.get_object(
            Bucket=settings.COS_BUCKET_NAME,
            Key=key,
        )
        md_body = b""
        while 1:
            chunk = res["Body"].read()
            if not chunk:
                break
            md_body += chunk
    except CosServiceError as e:
        logger.error(f"failed to get file from cos: {e}")
        return "", const.CodeEnum.COS_ERROR
    try:
        md = md_body.decode("utf-8")
    except UnicodeDecodeError as e:
        logger.error(f"failed to decode md: {e}. md: {md_body}")
        return "", const.CodeEnum.OPERATION_FAILED
    return md, const.CodeEnum.OK


def __save_md_to_cos(uid: str, nid: str, version: str, md: str) -> const.CodeEnum:
    settings = config.get_settings()
    cos_client, key, code = __cos_connect(settings, uid, nid, version)

    if code != const.CodeEnum.OK:
        return code

    # can raise error
    try:
        _ = cos_client.put_object(
            Bucket=settings.COS_BUCKET_NAME,
            Body=md.encode("utf-8"),
            Key=key,
            StorageClass='STANDARD',  # 'STANDARD'|'STANDARD_IA'|'ARCHIVE',
            EnableMD5=False,
            ContentType="text/markdown",
        )
    except CosServiceError as e:
        logger.error(f"failed to save file to cos: {e}")
        return const.CodeEnum.COS_ERROR
    return const.CodeEnum.OK


def __remove_md_from_cos(uid: str, nid: str, version: str) -> const.CodeEnum:
    settings = config.get_settings()
    cos_client, key, code = __cos_connect(settings, uid, nid, version, exist_ok=False)
    if code == const.CodeEnum.NODE_EXIST:
        try:
            _ = cos_client.delete_object(
                Bucket=settings.COS_BUCKET_NAME,
                Key=key
            )
        except CosServiceError as e:
            logger.error(f"failed to delete md from cos: {e}")
            return const.CodeEnum.COS_ERROR

    return const.CodeEnum.OK


def __remove_md_all_versions_from_cos(uid: str, nid: str) -> const.CodeEnum:
    settings = config.get_settings()
    cos_client, to_delete_dir = __get_client_and_key(settings, uid, nid)

    try:
        _ = cos_client.delete_object(
            Bucket=settings.COS_BUCKET_NAME,
            Key=to_delete_dir
        )
    except CosServiceError as e:
        logger.error(f"failed to delete md from cos: {e}")
        return const.CodeEnum.COS_ERROR

    return const.CodeEnum.OK


def __cos_connect(
        settings: BaseSettings,
        uid: str,
        nid: str,
        version: str,
        exist_ok: bool = False,
) -> Tuple["CosS3Client", str, const.CodeEnum]:
    cos_client, key = __get_client_and_key(settings, uid, nid, version)

    if not exist_ok:
        try:
            _ = cos_client.head_object(
                Bucket=settings.COS_BUCKET_NAME,
                Key=key
            )
            return cos_client, key, const.CodeEnum.NODE_EXIST
        except CosServiceError as e:
            if e.get_status_code() != 404:
                logger.error(f"failed to save md to cos: {e}")
                return cos_client, key, const.CodeEnum.COS_ERROR
    return cos_client, key, const.CodeEnum.OK


def __get_client_and_key(settings: BaseSettings, uid: str, nid: str, version: str = None) -> Tuple["CosS3Client", str]:
    cos_client = CosS3Client(
        CosConfig(
            Region=settings.COS_REGION,
            SecretId=settings.COS_SECRET_ID,
            SecretKey=settings.COS_SECRET_KEY,
            Token=None,
            Domain=None,
            Scheme='https',
        )
    )
    if version:
        key = f"mdHist/{uid}/{nid}/{version}.md"
    else:
        key = f"mdHist/{uid}/{nid}/"
    return cos_client, key



from datetime import datetime
from typing import TypedDict, List

from bson import ObjectId


# system notice from sender
class NoticeManagerDelivery(TypedDict):
    _id: ObjectId
    senderType: int  # user type, only admin and manager can send
    senderId: str
    title: str
    html: str
    snippet: str
    recipientType: int  # send to which user type, 0: batch, 1: all, 2: admin, 3: manager
    batchTypeIds: List[str]  # if recipient=batch, put user id here
    publishAt: datetime  # publish time
    scheduled: bool  # has been scheduled to sent to user


# system notice receiver
class NoticeSystem(TypedDict):
    _id: ObjectId
    senderId: str  # sender's uid
    recipientId: str  # recipient's uid
    noticeId: ObjectId  # notice _id
    read: bool  # has been read
    readTime: datetime  # read time

# event notice
# class EventRemind(TypedDict):
#     _id: ObjectId
#     event: int     # action type, 0: user reply, 1: user like, 2: user at, 3: user subscribe, 4: subscribe new post
#     senderId: str   # sender's uid
#     title: str
#     content: str
#     recipientId: str    # recipient's uid
#     read: bool      # has been read
#     readTime: datetime   # read time
